# Copyright 2021 The KCL Authors. All rights reserved.

import typing
from dataclasses import dataclass

import kclvm.config
import kclvm.kcl.error as kcl_error
import kclvm.kcl.info as kcl_info
import kclvm.kcl.ast as ast


KCLX_NODE_FIELD = "node"
BIN_OP_MAPPING = {
    ast.BinOp.Add: "Add",
    ast.BinOp.Sub: "Sub",
    ast.BinOp.Mul: "Mul",
    ast.BinOp.Div: "Div",
    ast.BinOp.Mod: "Mod",
    ast.BinOp.Pow: "Pow",
    ast.BinOp.LShift: "LShift",
    ast.BinOp.RShift: "RShift",
    ast.BinOp.BitOr: "BitOr",
    ast.BinOp.BitAnd: "BitAnd",
    ast.BinOp.BitXor: "BitXor",
    ast.BinOp.FloorDiv: "FloorDiv",
    ast.BinOp.As: "As",
    ast.BinOp.And: "And",
    ast.BinOp.Or: "Or",
}
AUG_OP_MAPPING = {
    ast.AugOp.Assign: "Assign",
    ast.AugOp.Add: "Add",
    ast.AugOp.Sub: "Sub",
    ast.AugOp.Mul: "Mul",
    ast.AugOp.Div: "Div",
    ast.AugOp.Mod: "Mod",
    ast.AugOp.Pow: "Pow",
    ast.AugOp.LShift: "LShift",
    ast.AugOp.RShift: "RShift",
    ast.AugOp.BitOr: "BitOr",
    ast.AugOp.BitXor: "BitXor",
    ast.AugOp.BitAnd: "BitAnd",
    ast.AugOp.FloorDiv: "FloorDiv",
}
UNARY_OP_MAPPING = {
    ast.UnaryOp.UAdd: "UAdd",
    ast.UnaryOp.USub: "USub",
    ast.UnaryOp.Invert: "Invert",
    ast.UnaryOp.Not: "Not",
}
CMP_OP_MAPPING = {
    ast.CmpOp.Eq: "Eq",
    ast.CmpOp.NotEq: "NotEq",
    ast.CmpOp.Lt: "Lt",
    ast.CmpOp.LtE: "LtE",
    ast.CmpOp.Gt: "Gt",
    ast.CmpOp.GtE: "GtE",
    ast.CmpOp.Is: "Is",
    ast.CmpOp.In: "In",
    ast.CmpOp.Not: "Not",
    ast.CmpOp.IsNot: "IsNot",
    ast.CmpOp.NotIn: "NotIn",
}
QUANT_OP_MAPPING = {
    ast.QuantOperation.ALL: "All",
    ast.QuantOperation.ANY: "Any",
    ast.QuantOperation.FILTER: "Filter",
    ast.QuantOperation.MAP: "Map",
}
CONFIG_ENTRY_OP_MAPPING = {
    ast.ConfigEntryOperation.UNION: "Union",
    ast.ConfigEntryOperation.OVERRIDE: "Override",
    ast.ConfigEntryOperation.INSERT: "Insert",
}
EXPR_CTX_MAPPING = {
    ast.ExprContext.LOAD: "Load",
    ast.ExprContext.STORE: "Store",
    ast.ExprContext.AUGLOAD: "Load",
    ast.ExprContext.AUGSTORE: "Store",
    ast.ExprContext.DEL: "Del",
}
OVERRIDE_ACTION_MAPPING = {
    ast.OverrideAction.CREATE_OR_UPDATE: "CreateOrUpdate",
    ast.OverrideAction.DELETE: "Delete",
}
TYPE_KCLX_ENUM_MAPPING = {
    # Stmt
    "TypeAliasStmt": "TypeAlias",
    "UnificationStmt": "Unification",
    "AssignStmt": "Assign",
    "AugAssignStmt": "AugAssign",
    "AssertStmt": "Assert",
    "IfStmt": "If",
    "ImportStmt": "Import",
    "SchemaIndexSignature": "SchemaIndexSignature",
    "SchemaAttr": "SchemaAttr",
    "SchemaStmt": "Schema",
    "RuleStmt": "Rule",
    # Expr
    "Identifier": "Identifier",
    "UnaryExpr": "Unary",
    "BinaryExpr": "Binary",
    "IfExpr": "If",
    "SelectorExpr": "Selector",
    "CallExpr": "Call",
    "ParenExpr": "Paren",
    "QuantExpr": "Quant",
    "ListExpr": "List",
    "ListIfItemExpr": "ListIfItem",
    "ListComp": "ListComp",
    "StarredExpr": "Starred",
    "DictComp": "DictComp",
    "ConfigIfEntryExpr": "ConfigIfEntry",
    "CompClause": "CompClause",
    "SchemaExpr": "Schema",
    "ConfigExpr": "Config",
    "ConfigEntry": "ConfigEntry",
    "CheckExpr": "Check",
    "LambdaExpr": "Lambda",
    "Decorator": "Decorator",
    "Subscript": "Subscript",
    "Keyword": "Keyword",
    "Arguments": "Arguments",
    "Compare": "Compare",
    "NumberLit": "NumberLit",
    "StringLit": "StringLit",
    "NameConstantLit": "NameConstantLit",
    "JoinedString": "JoinedString",
    "FormattedValue": "FormattedValue",
}
INIT_FILENAME = ""
INIT_POS = 1


@dataclass
class KCLxNode:
    filename: str
    line: int
    column: int
    end_line: int
    end_column: int


class BaseKCLxASTTransformer(ast.TreeWalker):
    @staticmethod
    def ast_meta_to_dict(t: ast.AST) -> dict:
        return KCLxNode(
            filename=t.filename or INIT_FILENAME,
            line=t.line or INIT_POS,
            column=t.column or INIT_POS,
            end_line=t.end_line or t.line or INIT_POS,
            end_column=t.end_column or t.end_column or INIT_POS,
        ).__dict__

    def get_node_name(self, t: ast.AST):
        return t.type

    def stmts(self, stmts: typing.List[ast.Stmt]):
        return [self.stmt(stmt) for stmt in stmts or []]

    def exprs(self, exprs: typing.List[ast.Expr], with_enum_name: bool = False):
        return [self.expr(expr, with_enum_name) for expr in exprs or []]

    def expr(self, expr: ast.Expr, with_enum_name: bool = False):
        expr_value = self.walk(expr) if expr else None
        if with_enum_name and expr_value and KCLX_NODE_FIELD in expr_value:
            expr_value[KCLX_NODE_FIELD] = {
                TYPE_KCLX_ENUM_MAPPING[expr._ast_type]: expr_value[KCLX_NODE_FIELD]
            }
        return expr_value

    def stmt(self, stmt: ast.Expr):
        return self.walk(stmt) if stmt else None


class KCLxASTTransformer(BaseKCLxASTTransformer):
    """TODO: Transform the Python KCL AST to the KCLx AST"""

    def walk_Module(self, t: ast.Module):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                "filename": t.filename,
                "pkg": t.pkg,
                "doc": t.doc,
                "name": t.name,
                "body": self.stmts(t.body),
                "comments": self.exprs(t.comments),
            }
        )
        return data

    def walk_TypeAliasStmt(self, t: ast.TypeAliasStmt):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    TYPE_KCLX_ENUM_MAPPING[t._ast_type]: {
                        "type_name": self.expr(t.type_name),
                        "type_value": {
                            KCLX_NODE_FIELD: t.type_value.plain_type_str,
                            **self.ast_meta_to_dict(t),
                        },
                    }
                }
            }
        )
        return data

    def walk_ExprStmt(self, t: ast.ExprStmt):
        data = self.ast_meta_to_dict(t)
        data.update({KCLX_NODE_FIELD: {"Expr": {"exprs": self.exprs(t.exprs, True)}}})
        return data

    def walk_UnificationStmt(self, t: ast.UnificationStmt):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    TYPE_KCLX_ENUM_MAPPING[t._ast_type]: {
                        "target": self.expr(t.target),
                        "value": self.expr(t.value),
                    }
                }
            }
        )
        return data

    def walk_AssignStmt(self, t: ast.AssignStmt):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    TYPE_KCLX_ENUM_MAPPING[t._ast_type]: {
                        "targets": self.exprs(t.targets),
                        "value": self.expr(t.value, True),
                        "type_annotation": {
                            KCLX_NODE_FIELD: t.type_annotation,
                            **self.ast_meta_to_dict(t),
                        }
                        if t.type_annotation
                        else None,
                    }
                }
            }
        )
        return data

    def walk_AugAssignStmt(self, t: ast.AugAssignStmt):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    TYPE_KCLX_ENUM_MAPPING[t._ast_type]: {
                        "target": self.expr(t.target),
                        "value": self.expr(t.value, True),
                        "op": AUG_OP_MAPPING[t.op],
                    }
                }
            }
        )
        return data

    def walk_AssertStmt(self, t: ast.AssertStmt):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    TYPE_KCLX_ENUM_MAPPING[t._ast_type]: {
                        "test": self.expr(t.test, True),
                        "if_cond": self.expr(t.if_cond, True),
                        "msg": self.expr(t.msg, True),
                    }
                }
            }
        )
        return data

    def walk_IfStmt(self, t: ast.IfStmt):
        data = self.ast_meta_to_dict(t)
        elif_stmt = None
        if t.elif_cond and t.elif_body:
            elif_stmt = ast.IfStmt()
            elif_stmt.set_ast_position(t)
            elif_stmt.cond = t.elif_cond[0]
            elif_stmt.body = t.elif_body[0]
            elif_stmt.elif_cond = t.elif_cond[1:]
            elif_stmt.elif_body = t.elif_body[1:]
            elif_stmt.else_body = t.else_body
        data.update(
            {
                KCLX_NODE_FIELD: {
                    TYPE_KCLX_ENUM_MAPPING[t._ast_type]: {
                        "cond": self.expr(t.cond, True),
                        "body": self.stmts(t.body),
                        "orelse": self.stmts([elif_stmt])
                        if elif_stmt
                        else self.stmts(t.else_body),
                    }
                }
            }
        )
        return data

    def walk_ImportStmt(self, t: ast.ImportStmt):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    TYPE_KCLX_ENUM_MAPPING[t._ast_type]: {
                        "path": t.path,
                        "rawpath": t.rawpath,
                        "name": t.name,
                        "asname": t.asname,
                    }
                }
            }
        )
        return data

    def walk_SchemaIndexSignature(self, t: ast.SchemaIndexSignature):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "key_name": t.key_name,
                    "value": self.expr(t.value, True),
                    "key_type": {
                        KCLX_NODE_FIELD: t.key_type,
                        **self.ast_meta_to_dict(t),
                    },
                    "value_type": {
                        KCLX_NODE_FIELD: t.value_type,
                        **self.ast_meta_to_dict(t.value_type_node),
                    },
                    "any_other": t.any_other,
                }
            }
        )
        return data

    def walk_SchemaAttr(self, t: ast.SchemaAttr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    TYPE_KCLX_ENUM_MAPPING[t._ast_type]: {
                        "doc": t.doc,
                        "name": {
                            KCLX_NODE_FIELD: t.name,
                            **self.ast_meta_to_dict(t),
                        },
                        "type_str": {
                            KCLX_NODE_FIELD: t.type_str,
                            **self.ast_meta_to_dict(t),
                        },
                        "value": self.expr(t.value, True),
                        "op": {"Bin": BIN_OP_MAPPING[t.op]}
                        if isinstance(t.op, ast.BinOp)
                        else (
                            {"Aug": AUG_OP_MAPPING[t.op]}
                            if isinstance(t.op, ast.AugOp)
                            else None
                        ),
                        "is_optional": t.is_optional,
                        "decorators": self.exprs(t.decorators),
                    }
                }
            }
        )
        return data

    def walk_SchemaStmt(self, t: ast.SchemaStmt):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    TYPE_KCLX_ENUM_MAPPING[t._ast_type]: {
                        "doc": t.doc,
                        "name": {
                            KCLX_NODE_FIELD: t.name,
                            **self.ast_meta_to_dict(t.name_node),
                        },
                        "parent_name": self.expr(t.parent_name),
                        "for_host_name": self.expr(t.for_host_name),
                        "is_mixin": t.is_mixin,
                        "is_protocol": t.is_protocol,
                        "args": self.expr(t.args),
                        "mixins": self.exprs(t.mixins),
                        "body": self.stmts(t.body),
                        "decorators": self.exprs(t.decorators),
                        "checks": self.exprs(t.checks),
                        "index_signature": self.stmt(t.index_signature),
                    }
                }
            }
        )
        return data

    def walk_RuleStmt(self, t: ast.RuleStmt):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    TYPE_KCLX_ENUM_MAPPING[t._ast_type]: {
                        "doc": t.doc,
                        "name": {
                            KCLX_NODE_FIELD: t.name,
                            **self.ast_meta_to_dict(t.name_node),
                        },
                        "parent_rules": self.exprs(t.parent_rules),
                        "for_host_name": self.expr(t.for_host_name),
                        "args": self.expr(t.args),
                        "decorators": self.exprs(t.decorators),
                        "checks": self.exprs(t.checks),
                    }
                }
            }
        )
        return data

    def walk_Identifier(self, t: ast.Identifier):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "names": t.names,
                    "pkgpath": t.names[0]
                    if t.names[0].startswith("@")
                    else (t.pkgpath or ""),
                    "ctx": EXPR_CTX_MAPPING.get(t.ctx, "Load"),
                }
            }
        )
        return data

    def walk_UnaryExpr(self, t: ast.UnaryExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "op": UNARY_OP_MAPPING[t.op],
                    "operand": self.expr(t.operand, True),
                }
            }
        )
        return data

    def walk_BinaryExpr(self, t: ast.BinaryExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "left": self.expr(t.left, True),
                    "op": {"Bin": BIN_OP_MAPPING[t.op]}
                    if isinstance(t.op, ast.BinOp)
                    else {"Cmp": CMP_OP_MAPPING[t.op]},
                    "right": self.expr(t.right, True),
                }
            }
        )
        return data

    def walk_IfExpr(self, t: ast.IfExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "body": self.expr(t.body, True),
                    "cond": self.expr(t.cond, True),
                    "orelse": self.expr(t.orelse, True),
                }
            }
        )
        return data

    def walk_SelectorExpr(self, t: ast.SelectorExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "value": self.expr(t.value, True),
                    "attr": self.expr(t.attr),
                    "ctx": EXPR_CTX_MAPPING[t.ctx],
                    "has_question": t.has_question,
                }
            }
        )
        return data

    def walk_CallExpr(self, t: ast.CallExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "func": self.expr(t.func, True),
                    "args": self.exprs(t.args, True),
                    "keywords": self.exprs(t.keywords),
                }
            }
        )
        return data

    def walk_ParenExpr(self, t: ast.ParenExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "expr": self.expr(t.expr, True),
                }
            }
        )
        return data

    def walk_QuantExpr(self, t: ast.QuantExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "target": self.expr(t.target, True),
                    "variables": self.exprs(t.variables),
                    "op": QUANT_OP_MAPPING[t.op],
                    "test": self.expr(t.test, True),
                    "if_cond": self.expr(t.if_cond, True),
                    "ctx": EXPR_CTX_MAPPING[t.ctx],
                }
            }
        )
        return data

    def walk_ListExpr(self, t: ast.ListExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "elts": self.exprs(t.elts, True),
                    "ctx": EXPR_CTX_MAPPING[t.ctx],
                }
            }
        )
        return data

    def walk_ListIfItemExpr(self, t: ast.ListIfItemExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "if_cond": self.expr(t.if_cond, True),
                    "exprs": self.exprs(t.exprs, True),
                    "orelse": self.expr(t.orelse, True),
                }
            }
        )
        return data

    def walk_ListComp(self, t: ast.ListComp):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "elt": self.expr(t.elt, True),
                    "generators": self.exprs(t.generators),
                }
            }
        )
        return data

    def walk_StarredExpr(self, t: ast.StarredExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "value": self.expr(t.value, True),
                    "ctx": EXPR_CTX_MAPPING[t.ctx],
                }
            }
        )
        return data

    def walk_DictComp(self, t: ast.DictComp):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "entry": {
                        "key": self.expr(t.key, True),
                        "value": self.expr(t.value, True),
                        "operation": CONFIG_ENTRY_OP_MAPPING[t.operation],
                        "insert_index": -1,
                    },
                    "generators": self.exprs(t.generators),
                }
            }
        )
        return data

    def walk_ConfigIfEntryExpr(self, t: ast.ConfigIfEntryExpr):
        data = self.ast_meta_to_dict(t)
        keys = []
        values = []
        for key, value, operation in zip(t.keys, t.values, t.operations):
            is_nest_key = isinstance(key, ast.Identifier) and len(key.names) > 1
            if is_nest_key:
                names = key.names
                key.names = [key.names[0]]
                for i, name in enumerate(names[1:][::-1]):
                    is_last_item = i == 0
                    name_node = ast.Identifier(
                        names=[name], line=key.line, column=key.column
                    )
                    name_node.filename = t.filename
                    entry_value = ast.ASTFactory.get_ast_configentry(
                        name_node,
                        value,
                        operation if is_last_item else ast.ConfigEntryOperation.UNION,
                        t.filename,
                    )
                    value = ast.ConfigExpr(line=key.line, column=key.column)
                    value.filename = t.filename
                    value.items.append(entry_value)
            keys.append(key)
            values.append(value)
        t.keys = keys
        t.values = values
        items = []
        for key, value, op in zip(t.keys, t.values, t.operations):
            items.append(
                ast.ConfigEntry(
                    line=key.line if key else value.line,
                    column=key.column if key else value.column,
                    key=key,
                    value=value,
                    operation=op,
                )
            )
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "if_cond": self.expr(t.if_cond, True),
                    "items": self.exprs(items),
                    "orelse": self.expr(t.orelse, True),
                }
            }
        )
        return data

    def walk_CompClause(self, t: ast.CompClause):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "targets": self.exprs(t.targets),
                    "iter": self.expr(t.iter, True),
                    "ifs": self.exprs(t.ifs, True),
                }
            }
        )
        return data

    def walk_SchemaExpr(self, t: ast.SchemaExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "name": self.expr(t.name),
                    "args": self.exprs(t.args, True),
                    "kwargs": self.exprs(t.kwargs),
                    "config": self.expr(t.config, True),
                }
            }
        )
        return data

    def walk_ConfigExpr(self, t: ast.ConfigExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "items": self.exprs(t.items),
                }
            }
        )
        return data

    def walk_ConfigEntry(self, t: ast.ConfigEntry):
        # Unpack the nest var form `a.b.c = 1` to `a: {b: {c = 1}}`
        is_nest_key = isinstance(t.key, ast.Identifier) and len(t.key.names) > 1
        if is_nest_key:
            names = t.key.names
            value = t.value
            t.key.names = [t.key.names[0]]
            for i, name in enumerate(names[1:][::-1]):
                is_last_item = i == 0
                name_node = ast.Identifier(
                    names=[name], line=t.key.line, column=t.key.column
                )
                name_node.filename = t.filename
                entry_value = ast.ASTFactory.get_ast_configentry(
                    name_node,
                    value,
                    t.operation if is_last_item else ast.ConfigEntryOperation.UNION,
                    t.filename,
                )
                value = ast.ConfigExpr(line=t.key.line, column=t.key.column)
                value.filename = t.filename
                value.items.append(entry_value)
            t.value = value
            t.operation = ast.ConfigEntryOperation.UNION
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "key": self.expr(t.key, True),
                    "value": self.expr(t.value, True),
                    "operation": CONFIG_ENTRY_OP_MAPPING[
                        ast.ConfigEntryOperation.UNION if is_nest_key else t.operation
                    ],
                    "insert_index": t.insert_index,
                }
            }
        )
        return data

    def walk_CheckExpr(self, t: ast.CheckExpr):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "test": self.expr(t.test, True),
                    "if_cond": self.expr(t.if_cond, True),
                    "msg": self.expr(t.msg, True),
                }
            }
        )
        return data

    def walk_LambdaExpr(self, t: ast.LambdaExpr):
        """ast.AST: LambdaExpr

        Parameters
        ----------
        - args: Optional[Arguments]
        - return_type_str: Optional[str]
        - return_type_node: Optional[Type]
        - body: List[Stmt]
        """
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "args": self.expr(t.args),
                    "return_type_str": t.return_type_str,
                    "body": self.stmts(t.body),
                }
            }
        )
        return data

    def walk_Decorator(self, t: ast.Decorator):
        name = self.expr(t.name, True)
        call_expr = self.expr(t.args)
        if call_expr:
            call_expr[KCLX_NODE_FIELD]["func"] = name
            return call_expr
        else:
            data = self.ast_meta_to_dict(t)
            data.update(
                {
                    KCLX_NODE_FIELD: {
                        "func": name,
                        "args": [],
                        "keywords": [],
                    }
                }
            )
            return data

    def walk_Subscript(self, t: ast.Subscript):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "value": self.expr(t.value, True),
                    "index": self.expr(t.index, True),
                    "lower": self.expr(t.lower, True),
                    "upper": self.expr(t.upper, True),
                    "step": self.expr(t.step, True),
                    "ctx": EXPR_CTX_MAPPING[t.ctx],
                    "has_question": t.has_question,
                }
            }
        )
        return data

    def walk_Keyword(self, t: ast.Keyword):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "arg": self.expr(t.arg),
                    "value": self.expr(t.value, True),
                }
            }
        )
        return data

    def walk_Arguments(self, t: ast.Arguments):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "args": self.exprs(t.args),
                    "defaults": self.exprs(t.defaults, True),
                    "type_annotation_list": [
                        {KCLX_NODE_FIELD: tpe_str, **self.ast_meta_to_dict(t)}
                        if tpe_str
                        else None
                        for tpe_str, tpe_node in zip(
                            t.type_annotation_list, t.type_annotation_node_list
                        )
                    ],
                }
            }
        )
        return data

    def walk_Compare(self, t: ast.Compare):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "left": self.expr(t.left, True),
                    "ops": [CMP_OP_MAPPING[op] for op in t.ops],
                    "comparators": self.exprs(t.comparators, True),
                }
            }
        )
        return data

    def walk_JoinedString(self, t: ast.JoinedString):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "is_long_string": t.is_long_string,
                    "values": self.exprs(t.values, True),
                    "raw_value": t.raw_value,
                }
            }
        )
        return data

    def walk_FormattedValue(self, t: ast.FormattedValue):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "is_long_string": t.is_long_string,
                    "value": self.expr(t.value, True),
                    "format_spec": t.format_spec,
                }
            }
        )
        return data

    def walk_NumberLit(self, t: ast.NumberLit):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "binary_suffix": t.binary_suffix,
                    "value": {"Int": t.value}
                    if isinstance(t.value, int)
                    else {"Float": t.value},
                }
            }
        )
        return data

    def walk_StringLit(self, t: ast.StringLit):
        data = self.ast_meta_to_dict(t)
        data.update(
            {
                KCLX_NODE_FIELD: {
                    "is_long_string": t.is_long_string,
                    "raw_value": t.raw_value or t.value,
                    "value": t.value,
                }
            }
        )
        return data

    def walk_NameConstantLit(self, t: ast.NameConstantLit):
        data = self.ast_meta_to_dict(t)
        value = "Undefined"
        value = "None" if t.value is None else value
        value = "True" if t.value is True else value
        value = "False" if t.value is False else value
        data.update({KCLX_NODE_FIELD: {"value": value}})
        return data

    def walk_Comment(self, t: ast.Comment):
        data = self.ast_meta_to_dict(t)
        data.update({KCLX_NODE_FIELD: {"text": t.text}})
        return data


def transform_ast_to_kclx_ast_json_str(program: ast.Program) -> str:
    check_number_lit_range(program)
    for pkgpath in program.pkgs:
        for i, module in enumerate(program.pkgs[pkgpath]):
            program.pkgs[pkgpath][i] = KCLxASTTransformer().walk_Module(module)
    return program.to_json(indent=None)


def check_number_lit_range(program: ast.Program):
    strict_range_check = kclvm.config.strict_range_check
    check_bit = 32 if strict_range_check else 64
    int_min = kcl_info.INT32_MIN if strict_range_check else kcl_info.INT64_MIN
    int_max = kcl_info.INT32_MAX if strict_range_check else kcl_info.INT64_MAX
    float_min = kcl_info.FLOAT32_MIN if strict_range_check else kcl_info.FLOAT64_MIN
    float_max = kcl_info.FLOAT32_MAX if strict_range_check else kcl_info.FLOAT64_MAX

    def walk_lit(t: ast.AST) -> typing.Optional[typing.Callable]:
        if isinstance(t, (ast.NumberLit)):
            numberLit = typing.cast(ast.NumberLit, t)
            value = numberLit.value

            if isinstance(value, int):
                if not (int_min <= value <= int_max):
                    kcl_error.report_exception(
                        err_type=kcl_error.ErrType.IntOverflow_TYPE,
                        file_msgs=[
                            kcl_error.ErrFileMsg(
                                filename=numberLit.filename, line_no=numberLit.line
                            )
                        ],
                        arg_msg=kcl_error.INT_OVER_FLOW_MSG.format(
                            str(value), check_bit
                        ),
                    )
            elif isinstance(value, float):
                abs_var = abs(value)
                if 0 < abs_var < float_min:
                    kcl_error.report_exception(
                        err_type=kcl_error.ErrType.FloatUnderflow_TYPE,
                        file_msgs=[
                            kcl_error.ErrFileMsg(
                                filename=numberLit.filename, line_no=numberLit.line
                            )
                        ],
                        arg_msg=kcl_error.FLOAT_UNDER_FLOW_MSG.format(
                            str(value), check_bit
                        ),
                    )
                elif abs_var > float_max:
                    kcl_error.report_exception(
                        err_type=kcl_error.ErrType.FloatOverflow_TYPE,
                        file_msgs=[
                            kcl_error.ErrFileMsg(
                                filename=numberLit.filename, line_no=numberLit.line
                            )
                        ],
                        arg_msg=kcl_error.FLOAT_OVER_FLOW_MSG.format(
                            str(value), check_bit
                        ),
                    )

        return walk_lit

    for pkgpath in program.pkgs:
        for i, module in enumerate(program.pkgs[pkgpath]):
            ast.WalkTree(module, walk_lit)



import io
import kclvm.kcl.error as kcl_error
import kclvm.tools.docs.templater as templater

# ---------------------------------------------------
# Constants
# ---------------------------------------------------


INVALID_FORMAT_MSG = "an unsupported format, expected {}"


# ---------------------------------------------------
# Factory
# ---------------------------------------------------


class IOFactory:
    def __init__(self):
        self._creators = {}

    def register_format(self, format: str, creator):
        self._creators[format] = creator

    def get(self, format: str, io: io.TextIOBase):
        creator = self._creators.get(format)
        if not creator:
            formats = ",".join(self._creators.keys())
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.CompileError_TYPE,
                arg_msg=INVALID_FORMAT_MSG.format(formats),
            )
        return creator(io)


class PathResolverFactory:
    def __init__(self):
        self._resolvers = {}

    def register_format(self, format: str, resolver):
        self._resolvers[format] = resolver

    def get(self, format: str, doc_name_formatter):
        resolver = self._resolvers.get(format)
        if not resolver:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.CompileError_TYPE,
                arg_msg=INVALID_FORMAT_MSG.format(",".join(self._resolvers.keys())),
            )
        return resolver(
            doc_name_formatter=doc_name_formatter,
            schema_section_render=templater.md_schema_section_render,
        )  # todo, get templater from factory


class DocEscaperFactory:
    def __init__(self):
        self._escapers = {}

    def register_format(self, format: str, escaper):
        self._escapers[format] = escaper

    def get(self, format: str):
        escaper = self._escapers.get(format)
        if not escaper:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.CompileError_TYPE,
                arg_msg=INVALID_FORMAT_MSG.format(",".join(self._escapers.keys())),
            )
        return escaper()



# Copyright 2021 The KCL Authors. All rights reserved.

import inspect

from dataclasses import dataclass, field
from typing import List, Dict, Union, Optional, Callable, TypeVar
from enum import IntEnum

import kclvm.kcl.error as kcl_error
import kclvm.api.object as obj
import kclvm.vm.code as vm_code
import kclvm.compiler.extension.builtin as builtin


# --------------------
# Type alias
# --------------------

SchemaKeyNativeType = Union[str, int, float]
SchemaKeyObjectType = Union[obj.KCLStringObject, obj.KCLIntObject, obj.KCLFloatObject]
SchemaKeyType = Union[SchemaKeyNativeType, SchemaKeyObjectType]
VirtualMachine = TypeVar("vm.VirtualMachine")


# ------------------------------------
# Schema attribute value place holder
# ------------------------------------


class ValuePlaceHolderPriority(IntEnum):
    """
    Value place holder override priority (Ascending)
    ------------------------------------------------
    1. base_default (default value of the base class schema attribute)
    2. base_templating (base schema writing general expression context)
    3. base_mixin (base class schema attribute)
    4. sub_default (the default value of the schema attribute of the subclass)
    5. sub_templating (subclass schema writing general expression context)
    6. sub_mixin (context of subclass schema)
    7. config (schema instantiation configuration value)
    """

    BASE_DEFAULT = 1
    BASE_TEMPLATING = 2
    BASE_MIXIN = 3
    SUB_DEFAULT = 4
    SUB_TEMPLATING = 5
    SUB_MIXIN = 6
    CONFIG = 7


@dataclass
class ValuePlaceHolder:
    """The real value of the schema attribute can be regarded
    as a placeholder for a non-literal expression before it is calculated

    Please note that all optional attribute checks, relaxed attr append,
    schema index signature and decorator execution and check block are executed
    after the calculation of the above process is completed.
    """

    name: str  # Attribute name
    priority: ValuePlaceHolderPriority
    types: List[str] = field(default_factory=list)  # Attribute type annotation list
    value: obj.KCLObject = None  # Actual value
    config_value: obj.KCLObject = None  # Config value
    codes: List[vm_code.SchemaBodyOpcodeFactory] = field(
        default_factory=list
    )  # VM eval byte codes


# --------------------
# Cache definition
# --------------------


class ValueCache:
    """Key-value pair cache
    - The key is a valid key type including `str`, `int` and `float`
    - The value is a KCL object
    """

    def __init__(self):
        self._cache: Dict[SchemaKeyNativeType, obj.KCLObject] = {}

    def __getitem__(self, key):
        return self.get(key)

    def __contains__(self, key):
        return key in self._cache

    def clear(self):
        self._cache.clear()

    def get(
        self, key: SchemaKeyNativeType, eval_func: Callable = None, *args, **kwargs
    ) -> Optional[obj.KCLObject]:
        """Get the value from the cache. When the parameter `eval_func` is
        provided and there is no such value in the cache, the value is calculated
        by calling `eval_func`
        """
        if key not in self._cache and eval_func and inspect.isfunction(eval_func):
            self._cache[key] = eval_func(*args, **kwargs)
        return self._cache.get(key)

    def set(self, key: SchemaKeyNativeType, value: obj.KCLObject):
        if not key:
            raise ValueError(f"Invalid key {key}")
        if not value or not isinstance(value, obj.KCLObject):
            raise ValueError(f"Invalid kcl object {value}")
        self._cache[key] = value


# --------------------
# Backtracking
# --------------------


@dataclass
class Backtracking:
    """Backtracking calculation set

    `__enter__` and `__exit__` function can be used as follows:

        backtracking = Backtracking()
        with backtracking.catch(name):
            assert Backtracking.tracking_level(name) == 1
            with backtracking.catch(name):
                assert Backtracking.tracking_level(name) == 2
            assert Backtracking.tracking_level(name) == 1
        assert Backtracking.tracking_level(name) == 0
    """

    _set: dict = field(default_factory=dict)
    _key: Optional[str] = None

    def __enter__(self):
        if self._key not in self._set:
            self._set[self._key] = 1
        else:
            self._set[self._key] += 1

    def __exit__(self, e_t, e_v, t_b):
        if self._key in self._set:
            self._set[self._key] -= 1
            if self._set[self._key] == 0:
                del self._set[self._key]

    def catch(self, key: str):
        self._key = key
        return self

    def is_backtracking(self, key: str):
        return key in self._set and self.tracking_level(key) > 0

    def tracking_level(self, key: str):
        return self._set.get(key, 0)

    def reset(self):
        self._set.clear()


# --------------------
# Schema eval context
# --------------------


@dataclass
class SchemaEvalContext:
    """Schema irrelevant order calculation context

    TODO: Need to be combined with configuration merging technology and
    move it into the compiler stage.
    """

    type_obj: obj.KCLSchemaTypeObject
    schema_obj: obj.KCLSchemaObject
    config: obj.KCLDictObject
    config_meta: dict
    args: List[obj.KCLObject]
    kwargs: List[obj.KWArg]
    vm: VirtualMachine
    cache: ValueCache = field(default_factory=ValueCache)

    place_holder_map: Dict[str, ValuePlaceHolder] = field(default_factory=lambda: None)
    backtracking: Backtracking = field(default_factory=Backtracking)

    def eval_reset(self):
        """Eval status reset to prevent reference interference"""
        self.cache.clear()
        self.backtracking.reset()

    # ----------------------------
    # Code to value place holders
    # ----------------------------

    def code_to_place_holder(
        self,
        code: vm_code.SchemaBodyOpcodeFactory,
        name: str,
        priority: ValuePlaceHolderPriority,
    ) -> ValuePlaceHolder:
        """Covert schema body code to value place holder"""
        assert (
            name
            and isinstance(name, str)
            and code
            and isinstance(code, vm_code.SchemaBodyOpcodeFactory)
        )
        place_holder = ValuePlaceHolder(
            name=name,
            priority=priority,
            types=self.schema_obj.get_attr_type(name),
            codes=[code],
            config_value=self.config.get(name),
        )
        return place_holder

    def get_all_place_holders(self) -> Dict[str, ValuePlaceHolder]:
        """Get all schema attribute value place holders"""
        self.place_holder_map = self.get_place_holders_by_type_obj(
            self.type_obj, {}, ValuePlaceHolderPriority.SUB_DEFAULT
        )
        return self.place_holder_map

    def get_place_holders_by_type_obj(
        self,
        type_obj: obj.KCLSchemaTypeObject,
        map_ref: dict,
        priority: ValuePlaceHolderPriority,
    ) -> Dict[str, ValuePlaceHolder]:
        """Get schema place holders using the schema type object"""
        if not type_obj or not isinstance(type_obj, obj.KCLSchemaTypeObject):
            return {}
        # Get the base schema place holders
        if type_obj.base:
            self.get_place_holders_by_type_obj(
                type_obj.base, map_ref, ValuePlaceHolderPriority.BASE_DEFAULT
            )
        code_factory = vm_code.SchemaBodyOpcodeFactory.build_from_codes(
            type_obj.func.instructions,
            type_obj.pkgpath,
            type_obj.name,
        )
        # Split opcodes into key-value pair `{attr_name}-{opcode_list}`
        codes = code_factory.split_to_schema_attr_codes()
        assert len(codes) == len(
            type_obj.attr_list
        ), f"{len(codes)} != {len(type_obj.attr_list)}"
        for i, code in enumerate(codes):
            place_holder = self.code_to_place_holder(
                code, type_obj.attr_list[i], priority + 1
            )
            if (
                place_holder
                and place_holder.name in map_ref
                and place_holder.priority >= map_ref[place_holder.name].priority
            ):
                # Schema attribute multi-override
                map_ref[place_holder.name].codes.extend(place_holder.codes)
            else:
                # First default value place holder
                map_ref[place_holder.name] = place_holder
        # Get the mixin place holders
        for mixin in type_obj.mixins or []:
            self.get_place_holders_by_type_obj(mixin, map_ref, priority + 2)
        return map_ref

    # ----------------------------
    # Back track
    # ----------------------------

    def back_track(self, name: str, place_holder: ValuePlaceHolder):
        if not name or not place_holder:
            return
        try:
            if self.backtracking.is_backtracking(name):
                pass
                """
                # Please note that KCL variable reference is not considered a self circular reference,
                # and the value may be overwritten such as `a = 1; a += 2` will be treated as `a = 1 + 2`
                raise kcl_error.KCLRuntimeError(
                    "RecursionError", None, None, None, f"Attribute '{name}' reference cycle"
                )
                """
            with self.backtracking.catch(name):
                # Number of variable backtracking
                level = self.backtracking.tracking_level(name)
                # Get the place holder code. Please note the negative index `-level`
                # because the later value place holders are placed at the end of the array
                code: vm_code.SchemaBodyOpcodeFactory = place_holder.codes[-level]
                # Run schema attribute place holder code
                self.vm_run_schema_code(code)
            # When the traceback ends, save the value to the cache
            self.cache.set(name, self.schema_obj.get(name))
        except IndexError:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.RecursionError_TYPE,
                arg_msg=f"Attribute '{name}' reference cycle",
            )

    # -------------------------------
    # Schema Attribute Getter/Setter
    # -------------------------------

    def set_value(self, name: SchemaKeyNativeType, value: obj.KCLObject):
        self.schema_obj.update({name: value})
        # If the attribute has only one place holder, its value is also the cache value
        if name in self.place_holder_map:
            if len(self.place_holder_map[name].codes) <= 1:
                self.cache.set(name, value)
            elif (
                self.place_holder_map[name].codes[-1].begin_index
                < self.vm.ctx.isp
                < self.place_holder_map[name].codes[-1].end_index
                and self.place_holder_map[name].codes[-1].schema_name
                == self.vm.ctx.name
            ):
                self.cache.set(name, value)

    def get_value(self, name: SchemaKeyNativeType) -> Optional[obj.KCLObject]:
        if not name:
            raise ValueError(f"Invalid name: {name}")
        # Deal in-place modify and return it self immediately
        if self.is_target_attr(name) and not self.backtracking.is_backtracking(name):
            return self.schema_obj.get(name)

        if name in self.cache:
            return self.cache[name]
        elif name in self.place_holder_map:
            self.back_track(name, self.place_holder_map.get(name))

        # 1. Load from schema self
        if name in self.schema_obj:
            value = self.schema_obj.get(name)
            return value

        # 2. Load from frame locals such as loop variables and schema args
        value = self.get_from_frame_locals(name)
        if value is not None:
            return value

        # 3. Load from globals
        value = self.get_from_frame_globals(name)
        if value is not None:
            return value

        # 4. Load from builtin
        built_obj_list = builtin.get_builtin_func_objects()
        for value in built_obj_list:
            if value.name == name:
                return value

        # 5. Load from pkg because the name may be a package variable that starts with `@` character
        pkgpath = name[1:] if name.startswith("@") else name
        if pkgpath in self.vm.state.modules:
            value = self.vm.state.modules[pkgpath]
            return value
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.EvaluationError_TYPE,
            arg_msg=f"'{name}' is not defined",
        )

    def is_target_attr(self, key: str) -> bool:
        if key not in self.place_holder_map:
            return False
        place_holders = self.place_holder_map.get(key)
        for code in place_holders.codes:
            if code.schema_name != self.vm.ctx.name:
                continue
            if code.begin_index <= self.vm.ctx.isp <= code.end_index:
                return True
        return False

    # -------------------------------------
    # VM schema body code runner
    # -------------------------------------

    def vm_run_schema_code(self, code: vm_code.SchemaBodyOpcodeFactory):
        if not code or not isinstance(code, vm_code.SchemaBodyOpcodeFactory):
            return
        codes_run: list = code.to_run_code_list()
        func = obj.KCLCompiledFunctionObject(
            name=code.schema_name,
            params=self.type_obj.func.params,
            names=self.type_obj.func.names,
            constants=self.type_obj.func.constants,
        )
        func.instructions = codes_run
        self.vm.push_frame_using_callable(
            code.pkgpath,
            func,
            [*self.args, self.config_meta, self.config, self.schema_obj]
            if self.args
            else [self.config_meta, self.config, self.schema_obj],
            self.kwargs,
            args_len=len(self.args),
        )
        # Run the schema compiled function body
        self.vm.run(run_current=True, ignore_nop=True)

    # -------------------------------------
    # VM frame locals/globals
    # -------------------------------------

    def get_from_frame_locals(self, key: str) -> Optional[obj.KCLDictObject]:
        """Get kcl object from vm frame locals dict"""

        # If the variable in the internal local scope:
        if self.vm.ctx.locals_list:
            index = len(self.vm.ctx.locals_list) - 1
            while index >= 0:
                if key in self.vm.ctx.locals_list[index]:
                    return self.vm.ctx.locals_list[index][key]
                index -= 1

        if self.key_is_in_frame_locals(key):
            return self.vm.ctx.locals[key]

        index = self.vm.frame_index
        # Search the local variables from the inside to the outside schema
        while index >= self.vm.frame_index:
            index -= 1
            if key in self.vm.frames[index].locals:
                return self.vm.frames[index].locals[key]
        return None

    def get_from_frame_globals(self, key: str) -> Optional[obj.KCLDictObject]:
        """Get kcl object from vm frame locals dict"""
        return self.vm.ctx.globals[key] if self.key_is_in_frame_globals(key) else None

    def key_is_in_frame_locals(self, key: str) -> bool:
        return key and isinstance(key, str) and key in self.vm.ctx.locals

    def key_is_in_frame_globals(self, key: str) -> bool:
        return key and isinstance(key, str) and key in self.vm.ctx.globals

    # ---------------
    # Static methods
    # ---------------

    @staticmethod
    def build_from_vm(
        vm: VirtualMachine,
        type_obj: obj.KCLSchemaTypeObject,
        schema_obj: obj.KCLSchemaObject,
        config: obj.KCLDictObject,
        config_meta: dict,
        args: List[obj.KCLObject],
        kwargs: List[obj.KWArg],
    ) -> "SchemaEvalContext":
        if not type_obj or not isinstance(type_obj, obj.KCLSchemaTypeObject):
            raise ValueError(f"Invalid kcl type object {type_obj}")
        context = SchemaEvalContext(
            type_obj=type_obj,
            schema_obj=schema_obj
            or obj.KCLSchemaObject(
                attrs={
                    obj.SCHEMA_SETTINGS_ATTR_NAME: obj.to_kcl_obj(type_obj.settings)
                },
                name=type_obj.name,
                runtime_type=type_obj.runtime_type,
            ),
            config=config,
            config_meta=config_meta,
            args=args,
            kwargs=kwargs,
            vm=vm,
            cache=ValueCache(),
        )
        return context



"""The `vm` file mainly contains the function `Run` which is used
# Copyright 2021 The KCL Authors. All rights reserved.
to execute the KCL bytecode obtained by the compiler module into
KCL result used to generate YAML/JSON data.
The KCL virtual machine receives a set of bytecode and uses its
own stack structure to store variables and the calculation results
of the variables, which includes `Frame` for context switching and
`VMState` for storing the import cache and executes the program
according to the corresponding opcode sequence and its operands.
When all the operation codes are executed, the entire KCL program
is also executed.
:copyright: Copyright 2020 The KCL Authors. All rights reserved.
"""

import os
import pathlib
import dataclasses
import typing

import kclvm.kcl.error as kcl_error
import kclvm.config
import kclvm.compiler.extension.builtin
import kclvm.compiler.extension.plugin
import kclvm.compiler.check.check_type

from kclvm.api.object import (
    KCLObject,
    KCLProgram,
    KCLResult,
    KCLCompiledFunctionObject,
    KCLModuleObject,
    KCLSchemaTypeObject,
    KWArg,
    to_kcl_obj,
)

from .code import Opcode, VM_OP_ACTIONS


@dataclasses.dataclass
class Frame:
    isp: int = 0
    name: str = None
    filename: str = None
    colno: int = 0
    lineno: int = 0
    pkgpath: str = None
    locals: dict = None
    globals: dict = None  # Global symbol and KCL object reference
    free_vars: list = None  # Free symbols
    locals_list: list = None
    codes: typing.List[int] = None

    def update_info(self, filename: str, lineno: int, colno: int):
        self.filename, self.lineno, self.colno = filename, lineno, colno

    def get_info(self) -> typing.Tuple[str, int, int]:
        return self.filename, self.lineno, self.colno


class VMState:
    def __init__(self):
        self.modules: typing.Dict[str, KCLModuleObject] = {}  # {pkgpath:module}
        self.globals_table: typing.Dict[
            str, typing.Dict[str, KCLObject]
        ] = {}  # {pkgpath:globals}


class VirtualMachine:
    """KCL Virtual Machine"""

    pkgpath_stack: typing.List[str] = []
    hook_output: str = ""

    @staticmethod
    def RunApp(app: KCLProgram, *, pkg: str = None) -> KCLResult:
        # Reset cache
        KCLSchemaTypeObject._eval_cache = {}
        VirtualMachine.pkgpath_stack = []
        VirtualMachine.hook_output = ""
        result = VirtualMachine(app).Run(pkg=pkg)
        if VirtualMachine.hook_output:
            result.m = to_kcl_obj(VirtualMachine.hook_output)
        return result

    def __init__(self, app: KCLProgram, state: VMState = None):
        super().__init__()

        self.app: KCLProgram = app
        self.state: VMState = state or VMState()

        self.names: typing.List[str] = []
        self.constants: typing.List[KCLObject] = []
        self.stack: typing.List[KCLObject] = []
        self.last_obj: typing.Optional[KCLObject] = None
        self.ctx: typing.Optional[Frame] = None
        self.frames: typing.List[Frame] = []
        self.frame_index = 1
        self.last_popped_frame = None
        self.sp: int = 0  # Stack Pointer

        self.cur_run_pkg: str = ""

        self.all_schema_types: typing.Dict[
            str, KCLSchemaTypeObject
        ] = {}  # [f"{pkgpath}.{name}": obj]

        self.lazy_eval_ctx = None

        self._reset(app.main)

    def Run(self, *, pkg: str = None) -> KCLResult:
        pkgpath = pkg if pkg else self.app.main
        if pkgpath in self.pkgpath_stack:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.RecursiveLoad_TYPE,
                arg_msg=kcl_error.RECURSIVE_LOADING_MODULE_MSG.format(
                    pkgpath, ", ".join(self.pkgpath_stack)
                ),
            )
        if pkg:
            self.pkgpath_stack.append(pkgpath)
        self._reset(pkgpath)
        return KCLResult(self.run(), self.get_filename())

    def define_schema_type(self, absname: str, schema_type: KCLSchemaTypeObject):
        if absname.startswith("@"):
            absname = absname[1:]
        self.all_schema_types[absname] = schema_type

    def find_schema_type(self, absname: str) -> typing.Optional[KCLSchemaTypeObject]:
        if absname.startswith("@"):
            absname = absname[1:]
        return self.all_schema_types.get(absname)

    def import_name(self, _code: int, arg: int) -> typing.Optional[KCLModuleObject]:
        self.pop()

        pkgpath = self.names[arg]
        asname = self.names[arg + 1]

        assert pkgpath and asname

        if pkgpath.startswith("@"):
            pkgpath = pkgpath[1:]
        if asname.startswith("@"):
            asname = asname[1:]

        if pkgpath in self.state.modules:
            # Read from module cache
            module = self.state.modules[pkgpath]

            self.update_local(asname, module)
            self.update_global(f"@{pkgpath}", module)
            self.push(module)

            return module

        if pkgpath in kclvm.compiler.extension.builtin.STANDARD_SYSTEM_MODULES:
            module = KCLModuleObject(
                name=pkgpath,
                asname=asname,
                value=kclvm.compiler.extension.builtin.get_system_module_func_objects(
                    pkgpath
                ),
            )

            self.update_local(asname, module)
            self.update_global(f"@{pkgpath}", module)
            self.push(module)

            # Module cache
            self.state.modules[pkgpath] = module
            self.state.globals_table[pkgpath] = {}

            return module

        if pkgpath.startswith(kclvm.compiler.extension.plugin.PLUGIN_MODULE_NAME):
            module = KCLModuleObject(
                name=pkgpath.replace(
                    kclvm.compiler.extension.plugin.PLUGIN_MODULE_NAME, ""
                ),
                asname=asname,
                value=kclvm.compiler.extension.plugin.get_plugin_func_objects(pkgpath),
            )

            self.update_local(asname, module)
            self.update_global(f"@{pkgpath}", module)
            self.push(module)

            self.state.modules[pkgpath] = module
            self.state.globals_table[pkgpath] = {}

            return module

        if pkgpath in self.app.pkgs:
            pkg_vm = VirtualMachine(self.app, self.state)
            pkg_vm.all_schema_types = self.all_schema_types

            result = pkg_vm.Run(pkg=pkgpath)

            module = KCLModuleObject(name=pkgpath, asname=asname, value=result.m)

            self.update_local(asname, module)
            self.update_global(f"@{pkgpath}", module)
            self.push(module)

            self.state.modules[pkgpath] = module
            self.state.globals_table[pkgpath] = result.m

            return module
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.CannotFindModule_TYPE,
            arg_msg=kcl_error.CANNOT_FIND_MODULE_MSG.format(
                pkgpath,
                str(
                    pathlib.Path(self.get_filename()).parent
                    / (pkgpath.replace(".", os.sep))
                ),
            ),
        )

    def _reset(self, pkg: str):
        assert pkg in self.app.pkgs

        self.cur_run_pkg = pkg
        bytecode = self.app.pkgs[pkg]

        self.names: typing.List[str] = bytecode.names
        self.constants: typing.List[KCLObject] = bytecode.constants
        self.stack: typing.List[KCLObject] = []
        self.last_obj: typing.Optional[KCLObject] = None
        self.ctx: typing.Optional[Frame] = None
        self.frames: typing.List[Frame] = [
            Frame(
                codes=bytecode.instructions,
                pkgpath=pkg,
                locals={},
                globals={},
                free_vars=[],
            )
        ]
        self.frame_index = 1
        self.last_popped_frame = None
        self.sp: int = 0  # Stack Pointer

    def run(self, run_current=False, ignore_nop=False):
        try:
            self.ctx = self.current_frame()
            while self.ctx.codes and self.ctx.isp < len(self.ctx.codes):
                codes = self.ctx.codes
                code = self.ctx.codes[self.ctx.isp]
                self.ctx.isp += 1
                # Skip the invalid opcode
                if code == Opcode.INVALID:
                    continue
                arg = None
                if Opcode.has_arg(code):
                    arg = (
                        codes[self.ctx.isp]
                        + (codes[self.ctx.isp + 1] << 8)
                        + (codes[self.ctx.isp + 2] << 16)
                    )
                    self.ctx.isp += 3
                info = self.ctx.codes[self.ctx.isp]
                info = typing.cast(tuple, info)
                self.ctx.isp += 1
                self.ctx.update_info(*info)
                action = VM_OP_ACTIONS.get(code)
                if ignore_nop and (code == Opcode.NOP or code == Opcode.SCHEMA_NOP):
                    continue
                if not action:
                    raise Exception(f"invalid opcode {code}")
                action(self, code, arg)
                if code == Opcode.RETURN_VALUE and run_current:
                    break
        except kcl_error.KCLException as err:
            if not err.lineno:
                filename, line, _ = self.get_info()
                errfile = err.pop_err_info()
                errfile.filename, errfile.line_no = filename, line or None
                err.append_err_info(errfile)
            raise err
        except Exception as err:
            if kclvm.config.debug and kclvm.config.verbose > 2:
                raise
            filename, lineno, _ = self.get_info()
            if isinstance(err, AttributeError):
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.AttributeError_Runtime_TYPE,
                    file_msgs=[kcl_error.ErrFileMsg(filename=filename, line_no=lineno)],
                    arg_msg=str(err),
                )
            if isinstance(err, RecursionError):
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.RecursionError_TYPE,
                    file_msgs=[kcl_error.ErrFileMsg(filename=filename, line_no=lineno)],
                    arg_msg=str(err),
                )
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.EvaluationError_TYPE,
                file_msgs=[kcl_error.ErrFileMsg(filename=filename, line_no=lineno)],
                arg_msg=str(err),
            )

        return self.ctx.globals

    def get_info(self, with_column=False) -> typing.Tuple[str, int, int]:
        filename, line, column = self.ctx.get_info()
        return filename, line, column if with_column else None

    def get_filename(self) -> str:
        filename, _, _ = self.ctx.get_info()
        return filename

    def load_const(self, index: int) -> KCLObject:
        cst = self.constants[index]
        self.push(cst)
        return cst

    def load_name(self, index: int) -> KCLObject:
        name = self.names[index]
        self.push(self.ctx.globals[name])
        return self.ctx.globals[name]

    def store_name(self, index: int):
        name = self.names[index]
        self.ctx.globals[name] = kclvm.compiler.check.check_type.check(
            self.stack_top(), *self.get_info()
        )

    def load_local(self, index: int):
        name = self.names[index]

        if self.ctx.locals_list:
            index = len(self.ctx.locals_list) - 1
            while index >= 0:
                if name in self.ctx.locals_list[index]:
                    self.push(self.ctx.locals_list[index][name])
                    return self.ctx.locals_list[index][name]
                index -= 1

        self.push(self.ctx.locals[name])
        return self.ctx.locals[name]

    def store_local(self, index: int):
        self.update_local(self.names[index], self.stack_top())

    def update_local(self, name: str, value: KCLObject, is_local_var: bool = False):
        if self.ctx.locals_list:
            index = len(self.ctx.locals_list) - 1
            while index >= 0:
                if name in self.ctx.locals_list[index] and not is_local_var:
                    self.ctx.locals_list[index][name] = value
                    return
                index -= 1

            self.ctx.locals_list[-1][name] = value
            return

        self.ctx.locals[name] = value

    def update_global(self, name: str, value: KCLObject):
        self.ctx.globals[name] = kclvm.compiler.check.check_type.check(
            value, *self.get_info()
        )

    def load_builtin(self, index: int):
        built_obj_list = kclvm.compiler.extension.builtin.get_builtin_func_objects()
        self.push(built_obj_list[index])

    def set_instruction_pointer(self, index: int):
        self.ctx.isp = int(index)

    def stack_top(self) -> KCLObject:
        return self.stack[-1]

    def current_frame(self) -> Frame:
        return self.frames[-1]

    def push_frame(self, frame: Frame, names=None, constants=None):
        self.frames.append(frame)
        self.ctx = self.frames[-1]
        self.frame_index += 1

    def push_frame_using_callable(
        self,
        pkgpath: str,
        func: KCLCompiledFunctionObject,
        args: typing.List[KCLObject],
        kwargs: typing.List[KWArg],
        args_len: int = 0,
    ):
        assert isinstance(func, KCLCompiledFunctionObject)

        filename, line, column = self.get_info()

        ctx_globals = self.frames[-1].globals
        if pkgpath in self.state.globals_table:
            ctx_globals = self.state.globals_table[pkgpath]
            self.cur_run_pkg = pkgpath

        self.push_frame(
            Frame(
                name=func.name,
                codes=func.instructions,
                pkgpath=pkgpath,
                locals={},
                globals=ctx_globals,
                free_vars=[],
            ),
            func.names,
            func.constants,
        )
        arg_index = 0
        for arg in args[args_len:] or []:
            self.push(arg)
        for default_arg in func.params:
            if default_arg.value:
                self.update_local(default_arg.name, default_arg.value)
        # args[3:] - schema args
        for arg in args[:args_len]:
            name = func.params[arg_index].name
            self.update_local(name, arg)
            arg_index += 1
        for kwarg in kwargs or []:
            name = kwarg.name.value
            if name not in [p.name for p in func.params]:
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.EvaluationError_TYPE,
                    file_msgs=[
                        kcl_error.ErrFileMsg(
                            filename=filename, line_no=line, col_no=column
                        )
                    ],
                    arg_msg=f"schema arguments got an unexpected keyword argument '{name}'",
                )
            self.update_local(name, kwarg.value)

    def push_frame_locals(self):
        self.ctx.locals_list = self.ctx.locals_list or []
        self.ctx.locals_list.append({})

    def pop_frame_locals(self):
        self.ctx.locals_list.pop()

    def pop_frame(self) -> Frame:
        self.last_popped_frame = self.frames.pop()
        self.ctx = self.frames[-1]
        self.frame_index -= 1
        return self.last_popped_frame

    def push_function(self, index: int):
        func_obj = self.constants[index]
        self.push(func_obj)

    def push(self, obj: KCLObject):
        self.stack.append(obj)

    def pop(self):
        self.last_obj = self.stack.pop()
        return self.last_obj

    def peek(self) -> KCLObject:
        return self.stack[-1]

    def peek_nth(self, index: int) -> KCLObject:
        """View the Nth top item on the stack from index 0"""
        return self.stack[-index]

    def top(self) -> KCLObject:
        """Get the top of the VM stack"""
        return self.stack[-1]

    def set_top(self, obj: KCLObject):
        """Set the top of the VM stack"""
        self.stack[-1] = obj

    def clear(self):
        self.stack.clear()

    def last_popped_obj(self) -> KCLObject:
        return self.last_obj

    def debug_stack(self, idx: int, at: int = 0):
        if at > 0:
            print(f"vm.stack[{-1-idx}]={self.stack[-1-idx]} # at({at})")
        else:
            print(f"vm.stack[{-1-idx}]={self.stack[-1-idx]}")

    def debug_locals(self, _arg: int, at: int = 0):
        if at > 0:
            print(f"vm.ctx.locals={self.ctx.locals} # at({at})")
        else:
            print(f"vm.ctx.locals={self.ctx.locals}")

    def debug_globals(self, _arg: int, at: int = 0):
        if at > 0:
            print(f"vm.ctx.globals={self.ctx.globals} # at({at})")
        else:
            print(f"vm.ctx.globals={self.ctx.globals}")

    def debug_names(self, _arg: int, at: int = 0):
        if at > 0:
            print(f"vm.names={self.names} # at({at})")
        else:
            print(f"vm.names={self.names}")


def Run(app: KCLProgram, *, pkg: str = None) -> KCLResult:
    return VirtualMachine.RunApp(app, pkg=pkg)



from typing import Any
from abc import abstractmethod

import kclvm.kcl.ast as ast


class BaseChecker(ast.TreeWalker):
    # checker name (you may reuse an existing one)
    name: str = ""
    # options level (0 will be displaying in --help, 1 in --long-help)
    level = 1
    # ordered list of options to control the checker behaviour
    options: Any = {}
    # messages constant to display
    MSGS: Any = {}
    # messages issued by this checker
    msgs: Any = []
    # mark this checker as enabled or not.
    enabled: bool = True

    def __init__(self, linter=None) -> None:
        """
        checker instances should have the linter as argument

        :param linter: is an object implementing KCLLinter.
        """
        if self.name is not None:
            self.name = self.name.lower()
        self.linter = linter
        self.options = linter.config if linter else None

    def __eq__(self, other) -> bool:
        return self.name == other.name and self.linter == other.linter

    def generic_walk(self, t: ast.AST) -> None:
        """Called if no explicit walker function exists for a node."""
        for _, value in ast.iter_fields(t):
            if isinstance(value, list):
                for v in value:
                    # IfStmt.elif_body: List[List[Stmt]]
                    if isinstance(v, list):
                        for v1 in v:
                            self.walk(v1)
                    if isinstance(v, ast.AST):
                        self.walk(v)
            elif isinstance(value, ast.AST):
                self.walk(value)

    def get_node_name(self, t: ast.AST) -> str:
        """Get the ast.AST node name"""
        assert isinstance(t, ast.AST)
        return t.type

    @abstractmethod
    def check(self, prog: ast.Program, code: str):
        """Should be overridden by subclass"""
        raise NotImplementedError()



# Copyright 2021 The KCL Authors. All rights reserved.

import json
import sys
import traceback

from typing import List, Dict, Callable

import google.protobuf.json_format as json_format

import kclvm.kcl.error as kcl_error

from .gpyrpc_pb2 import (
    Error,
    Request,
    Response,
    ListMethod_Result,
    Ping_Args,
    Ping_Result,
)

# https://developers.google.com/protocol-buffers/docs/reference/google.protobuf
# https://developers.google.com/protocol-buffers/docs/reference/python-generated
# https://github.com/protocolbuffers/protobuf/tree/master/src/google/protobuf

_rpc_method_table: Dict[str, Callable[[Request], Response]] = {}


def _gpyrpcCallProxy(method: str, jsonpb_Request: str) -> str:
    assert method
    assert jsonpb_Request

    try:
        fn = _rpc_method_table[method]
        assert fn

        req = Request()
        json_format.Parse(jsonpb_Request, req)

        resp = Response()
        fn_result = fn(req)
        resp.CopyFrom(fn_result)

        jsonpb_Response: str = json_format.MessageToJson(
            resp, including_default_value_fields=True, preserving_proto_field_name=True
        )
        return jsonpb_Response

    except kcl_error.KCLException as err:
        errMessage = f"{err}"
        resp = Response(err=Error(message=errMessage))
        return json_format.MessageToJson(
            resp, including_default_value_fields=True, preserving_proto_field_name=True
        )
    except OSError as err:
        errMessage = f"OSError: {err}"
        resp = Response(err=Error(message=errMessage))
        return json_format.MessageToJson(
            resp, including_default_value_fields=True, preserving_proto_field_name=True
        )
    except AssertionError as err:
        ex_type, ex_val, ex_stack = sys.exc_info()
        tb_info = traceback.extract_tb(ex_stack)[-1]

        filename = tb_info.filename
        line = tb_info.lineno

        errMessage = f"AssertionError: {err}"
        resp = Response(err=Error(message=errMessage, filename=filename, line=line))
        return json_format.MessageToJson(
            resp, including_default_value_fields=True, preserving_proto_field_name=True
        )
    except Exception as err:
        errMessage = f"Exception: Internal Error! Please report a bug to us: method={method}, err={err}, stack trace={traceback.format_exc()}"
        resp = Response(err=Error(message=errMessage))
        return json_format.MessageToJson(
            resp, including_default_value_fields=True, preserving_proto_field_name=True
        )


def gpyrpcCallProxy(method: str, jsonRequest: str) -> str:
    if not method:
        errMessage = "method is empty"
        resp = Response(err=Error(message=errMessage))
        return json_format.MessageToJson(
            resp, including_default_value_fields=True, preserving_proto_field_name=True
        )

    if method not in _rpc_method_table:
        errMessage = f"method '{method}' not found"
        resp = Response(err=Error(message=errMessage))
        return json_format.MessageToJson(
            resp, including_default_value_fields=True, preserving_proto_field_name=True
        )
        resp: Response = {"error": f"method '{method}' not found"}
        return str(json.dumps(resp))

    return _gpyrpcCallProxy(method, jsonRequest)


def RegisterMethod(method: str, fn: Callable[[Request], Response]):
    assert method
    assert fn

    assert method not in _rpc_method_table
    _rpc_method_table[method] = fn


# args: gpyrpc.Ping_Args
# result: gpyrpc.Ping_Result
def _gpyrpcPing(req: Request) -> Response:
    args = Ping_Args()
    req.args.Unpack(args)

    resp = Response()
    resp.result.Pack(Ping_Result(value=args.value))
    return resp


# args: gpyrpc.ListMethod_Args
# result: gpyrpc.ListMethod_Result
def _gpyrpcListMethod(req: Request) -> Response:
    keys: List[str] = list(sorted(_rpc_method_table))

    resp = Response()
    resp.result.Pack(ListMethod_Result(method_name_list=keys))

    return resp


RegisterMethod("gpyrpc.BuiltinService.Ping", _gpyrpcPing)
RegisterMethod("gpyrpc.BuiltinService.ListMethod", _gpyrpcListMethod)



from .builtin import (
    kcl_builtin,
    BUILTIN_FUNCTIONS,
    STANDARD_SYSTEM_MODULES,
    get_builtin_func_objects,
    get_system_module_func_objects,
    get_system_module_members,
)

__all__ = [
    "kcl_builtin",
    "BUILTIN_FUNCTIONS",
    "STANDARD_SYSTEM_MODULES",
    "get_builtin_func_objects",
    "get_system_module_func_objects",
    "get_system_module_members",
]



"""
The `kcl_err_template` file mainly contains some string templates to organize the structure of the KCL error message.

The KCL error message comes from kcl_err_msg.py

KCLErrMsgTemplate: An abstract class, which specifies the methods required by KCL error message template classes.

KCLErrMsgTemplateDefault: The default KCL template class implementation.

Provide four file message templates in KCLErrMsgTemplateDefault:

    get_hint_msg_common:
----------------------------------------------------------------------------
        KCL Compile Error[E2L23] : A complie error occurs during compiling
        ---> File /main.k:2:14
        2 |    name: str
                     14 ^  -> Failure
        An error occurs
----------------------------------------------------------------------------

    get_hint_msg_tiny:
----------------------------------------------------------------------------
        KCL Compile Error[E2L23] : A compile error occurs during compiling
        An error occurs
----------------------------------------------------------------------------

    get_hint_msg_summary:
----------------------------------------------------------------------------
        KCL Compile Error[E2L23] : A complie error occurs during compiling
        ---> File /main.k:2
        2 |    name: str -> Failure
        An error occurs
----------------------------------------------------------------------------

    get_hint_msg_detail:
----------------------------------------------------------------------------
        KCL Compile Error[E2L23] : A complie error occurs during compiling
        ---> File /main.k:2:5
        2 |    name: str
             5 ^^^^^^^^^  -> Failure
        An error occurs
----------------------------------------------------------------------------

Provide one error message templates in KCLErrMsgTemplateDefault:

    err_msg_template:
----------------------------------------------------------------------------
    KCL Compile Error[E2L23] : A compile error occurs during compiling
----------------------------------------------------------------------------

color_txt_err_msg(), color_txt_err_file_msg() in KCLErrMsgTemplateDefault
will call methods provided in kcl_err_template.py to highlight some fields
in the kcl error message

:copyright: Copyright 2020 The KCL Authors. All rights reserved.
"""
import os
import typing
import threading

from enum import unique, Enum
from pathlib import PosixPath
from string import Template
from abc import ABCMeta, abstractmethod
from kclvm.kcl.error.kcl_err_theme import (
    ColorOption,
    KCLErrMsgTheme,
    KCLErrMsgThemeDefault,
)
from kclvm.kcl.error.kcl_err_msg import KCL_ERR_MSG, ErrEwcode
from kclvm.internal.util import PreCheck, CheckRules, PostCheck, check_utils

DEFAULT_MSG_2 = "Failure"


@unique
class ErrLevel(Enum):
    SERIOUS = 1
    ORDINARY = 2


@unique
class MsgId(Enum):
    ERR_TYPE = 1
    ERR_CATE = 2
    EWCODE = 3
    MSG_1 = 4
    MSG_2 = 5

    FILENAME = 8
    SRC_CODE_LINE = 9
    LINE_NO = 10
    COL_NO = 11
    ERR_ARG = 12
    MARK = 13


@unique
class ErrMsgFmt(Enum):
    COLOR_TXT = 1
    NO_COLOR_TXT = 2


class FileCache:
    """File cache to store the filename and code mapping"""

    _file_cache: typing.Dict[str, str] = {}
    _lock = threading.RLock()

    @staticmethod
    def clear():
        FileCache._file_cache.clear()

    @staticmethod
    def put(file: str, code: str):
        FileCache._lock.acquire()
        FileCache._file_cache[file] = code
        FileCache._lock.release()

    @staticmethod
    def get(file: str) -> str:
        return FileCache._file_cache.get(file, "")


UP_ARROW = "^"
WAVY = "~"

MARK_SYMBOL = {ErrLevel.SERIOUS: UP_ARROW, ErrLevel.ORDINARY: WAVY}

TAB = "    "
COLON = ":"
COMMA = ", "
WHITE_SPACE = " "
SPARATOER = " |"
NEW_LINE = "\n"
HINT_MSG = "{}"
FILE_PREFIX = "---> File "
MSG_PREFIX = "KCL "


class KCLErrMsgTemplate(metaclass=ABCMeta):
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, ErrMsgFmt)), "fmt")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, KCLErrMsgTheme)), "theme")
    def __init__(
        self,
        fmt: ErrMsgFmt = ErrMsgFmt.NO_COLOR_TXT,
        theme: KCLErrMsgTheme = KCLErrMsgThemeDefault(),
    ):
        self.fmt = fmt
        self.theme = theme

    @abstractmethod
    def get_hint_msg_common(
        self,
        filename: str,
        line_no: int,
        col_no: int,
        src_code_line: str,
        indent_count: int = 0,
        mark: str = MARK_SYMBOL[ErrLevel.SERIOUS],
    ) -> str:
        pass

    @abstractmethod
    def get_hint_msg_tiny(self, filename: str, indent_count: int = 0) -> str:
        pass

    @abstractmethod
    def get_hint_msg_summary(
        self, filename: str, line_no: int, src_code_line: str, indent_count: int = 0
    ) -> str:
        pass

    @abstractmethod
    def get_hint_msg_detail(
        self,
        filename: str,
        line_no: int,
        col_no: int,
        end_col_no: int,
        src_code_line: str,
        indent_count: int = 0,
        mark: str = MARK_SYMBOL[ErrLevel.SERIOUS],
    ) -> str:
        pass

    @abstractmethod
    def err_msg_template(
        self, err_type: str, err_cate: str, ewcode_fmt: str, msg_1: str
    ):
        pass

    @abstractmethod
    def color_txt_err_file_msg(self, mark_level: ErrLevel):
        pass

    @abstractmethod
    def color_txt_err_msg(self, err_type: str, err_cate: str):
        pass


class KCLErrMsgTemplateDefault(KCLErrMsgTemplate):
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, ErrMsgFmt)), "fmt")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, KCLErrMsgTheme)), "theme")
    def __init__(
        self,
        fmt: ErrMsgFmt = ErrMsgFmt.NO_COLOR_TXT,
        theme: KCLErrMsgTheme = KCLErrMsgThemeDefault(),
    ):
        super().__init__(fmt, theme)
        self.fmt = fmt
        self.theme = theme

        self.ErrMsgLables = {
            MsgId.ERR_TYPE: "${err_type}",
            MsgId.ERR_CATE: "${err_cate}",
            MsgId.EWCODE: "${ewcode}",
            MsgId.MSG_1: "${msg_1}",
            MsgId.MSG_2: "${msg_2}",
            MsgId.FILENAME: "${filename}",
            MsgId.SRC_CODE_LINE: "${src_code_line}",
            MsgId.LINE_NO: "${line_no}",
            MsgId.COL_NO: "${col_no}",
            MsgId.ERR_ARG: "${err_arg}",
            MsgId.MARK: "${mark}",
        }
        self.FILENAME = self.ErrMsgLables[MsgId.FILENAME]
        self.LINE_NO = self.ErrMsgLables[MsgId.LINE_NO]
        self.COL_NO = self.ErrMsgLables[MsgId.COL_NO]
        self.MARK = self.ErrMsgLables[MsgId.MARK]

        self.EWCODE = self.ErrMsgLables[MsgId.EWCODE]
        self.ERR_TYPE = self.ErrMsgLables[MsgId.ERR_TYPE]
        self.ERR_CATE = self.ErrMsgLables[MsgId.ERR_CATE]
        self.MSG_1 = self.ErrMsgLables[MsgId.MSG_1]
        self.MSG_2 = self.ErrMsgLables[MsgId.MSG_2]

        self.SRC_CODE_LINE = self.ErrMsgLables[MsgId.SRC_CODE_LINE]
        self.ERR_ARG = self.ErrMsgLables[MsgId.ERR_ARG]

    def clean_color(self):
        self.FILENAME = self.ErrMsgLables[MsgId.FILENAME]
        self.LINE_NO = self.ErrMsgLables[MsgId.LINE_NO]
        self.COL_NO = self.ErrMsgLables[MsgId.COL_NO]
        self.MARK = self.ErrMsgLables[MsgId.MARK]

        self.EWCODE = self.ErrMsgLables[MsgId.EWCODE]
        self.ERR_TYPE = self.ErrMsgLables[MsgId.ERR_TYPE]
        self.ERR_CATE = self.ErrMsgLables[MsgId.ERR_CATE]

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "filename")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, int)), "line_no")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, int)), "col_no")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "src_code_line")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, int)), "indent_count")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "mark")
    @PostCheck(lambda v: CheckRules.check_type_not_none(v, str))
    def get_hint_msg_common(
        self,
        filename: str,
        line_no: int,
        col_no: int,
        src_code_line: str,
        indent_count: int = 0,
        mark: str = MARK_SYMBOL[ErrLevel.SERIOUS],
    ) -> str:
        HINT_MSG_COMMON = (
            NEW_LINE
            + TAB * indent_count
            + FILE_PREFIX
            + self.FILENAME
            + COLON
            + self.LINE_NO
            + COLON
            + self.COL_NO
            + NEW_LINE
            + TAB * indent_count
            + self.LINE_NO
            + SPARATOER
            + self.SRC_CODE_LINE
            + NEW_LINE
            + TAB * indent_count
            + (self.COL_NO + WHITE_SPACE).rjust(
                len(str(line_no)) + 1 + col_no + len(self.COL_NO) - len(str(col_no)),
                WHITE_SPACE,
            )
            + self.MARK
            + WHITE_SPACE
        )
        hint_msg_args = {
            "filename": filename,
            "line_no": line_no,
            "col_no": col_no,
            "src_code_line": src_code_line,
            "mark": mark,
        }
        return Template(HINT_MSG_COMMON).substitute(hint_msg_args)

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "filename")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, int)), "indent_count")
    @PostCheck(lambda v: CheckRules.check_type_not_none(v, str))
    def get_hint_msg_tiny(self, filename: str, indent_count: int = 0) -> str:
        HINT_MSG_TINY = NEW_LINE + TAB * indent_count + FILE_PREFIX + self.FILENAME
        hint_msg_args = {"filename": filename}
        return Template(HINT_MSG_TINY).substitute(hint_msg_args)

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "filename")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, int)), "line_no")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "src_code_line")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, int)), "indent_count")
    @PostCheck(lambda v: CheckRules.check_type_not_none(v, str))
    def get_hint_msg_summary(
        self, filename: str, line_no: int, src_code_line: str, indent_count: int = 0
    ) -> str:
        HINT_MSG_SUMMARY = (
            NEW_LINE
            + TAB * indent_count
            + FILE_PREFIX
            + self.FILENAME
            + COLON
            + self.LINE_NO
            + NEW_LINE
            + TAB * indent_count
            + self.LINE_NO
            + SPARATOER
            + self.SRC_CODE_LINE
        )
        hint_msg_args = {
            "filename": filename,
            "line_no": line_no,
            "src_code_line": src_code_line,
        }
        return Template(HINT_MSG_SUMMARY).substitute(hint_msg_args)

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "filename")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, int)), "line_no")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, int)), "col_no")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, int)), "end_col_no")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "src_code_line")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, int)), "indent_count")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "mark")
    @PostCheck(lambda v: CheckRules.check_type_not_none(v, str))
    def get_hint_msg_detail(
        self,
        filename: str,
        line_no: int,
        col_no: int,
        end_col_no: int,
        src_code_line: str,
        indent_count: int = 0,
        mark: str = MARK_SYMBOL[ErrLevel.SERIOUS],
    ) -> str:
        HINT_MSG_DETAIL = (
            NEW_LINE
            + TAB * indent_count
            + FILE_PREFIX
            + self.FILENAME
            + COLON
            + self.LINE_NO
            + COLON
            + self.COL_NO
            + NEW_LINE
            + TAB * indent_count
            + self.LINE_NO
            + SPARATOER
            + self.SRC_CODE_LINE
            + NEW_LINE
            + TAB * indent_count
            + (self.COL_NO + WHITE_SPACE).rjust(
                len(str(line_no)) + 1 + col_no + len(self.COL_NO) - len(str(col_no)),
                WHITE_SPACE,
            )
            + self.MARK * (end_col_no - col_no)
            + WHITE_SPACE
        )
        hint_msg_args = {
            "filename": filename,
            "line_no": line_no,
            "col_no": col_no,
            "src_code_line": src_code_line,
            "mark": mark,
        }
        return Template(HINT_MSG_DETAIL).substitute(hint_msg_args)

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "err_type")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "err_cate")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "ewcode_fmt")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "msg_1")
    @PostCheck(lambda v: CheckRules.check_type_not_none(v, str))
    def err_msg_template(
        self, err_type: str, err_cate: str, ewcode_fmt: str, msg_1: str
    ):
        SIMPLE_TEMPLATE_DOC = (
            MSG_PREFIX
            + self.ERR_TYPE
            + WHITE_SPACE
            + self.ERR_CATE
            + self.EWCODE
            + WHITE_SPACE
            + COLON
            + WHITE_SPACE
            + self.MSG_1
        )
        simple_args = {
            "err_type": err_type,
            "err_cate": err_cate,
            "ewcode": ewcode_fmt,
            "msg_1": msg_1,
        }
        return Template(SIMPLE_TEMPLATE_DOC).substitute(simple_args)

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, ErrLevel)), "mark_level")
    def color_txt_err_file_msg(self, mark_level: ErrLevel):
        self.MARK = self.theme.color_mark(self.MARK, mark_level.value)
        self.FILENAME = self.theme.color(self.FILENAME, ColorOption.FILE_NAME)
        self.LINE_NO = self.theme.color(self.LINE_NO, ColorOption.LINE_COLUMN)
        self.COL_NO = self.theme.color(self.COL_NO, ColorOption.LINE_COLUMN)

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "err_type")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "err_cate")
    def color_txt_err_msg(self, err_type: str, err_cate: str):
        self.EWCODE = self.theme.color(self.EWCODE, ColorOption.EWCODE)
        self.ERR_TYPE = self.color_err_msg(slot=self.ERR_TYPE, msg=err_type)
        self.ERR_CATE = self.color_err_msg(slot=self.ERR_CATE, msg=err_cate)

    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "msg")
    @PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "slot")
    @PostCheck(lambda v: CheckRules.check_type_not_none(v, str))
    def color_err_msg(self, slot: str, msg: str):
        msg_theme_map: dict = {
            KCL_ERR_MSG.get_err_msg_by_errid(ErrEwcode.KCLError_Ew): ColorOption.ERROR,
            KCL_ERR_MSG.get_err_msg_by_errid(
                ErrEwcode.KCLWarning_Ew
            ): ColorOption.WARNING,
            KCL_ERR_MSG.get_err_msg_by_errid(
                ErrEwcode.KCLSyntaxException_Ew
            ): ColorOption.SYNTAX,
            KCL_ERR_MSG.get_err_msg_by_errid(
                ErrEwcode.KCLCompileException_Ew
            ): ColorOption.COMPLIE,
            KCL_ERR_MSG.get_err_msg_by_errid(
                ErrEwcode.KCLRuntimeException_Ew
            ): ColorOption.RUNTIME,
            KCL_ERR_MSG.get_err_msg_by_errid(
                ErrEwcode.KCLException_Ew
            ): ColorOption.ERROR,
        }
        try:
            return self.theme.color(slot, msg_theme_map[msg])
        except KeyError:
            check_utils.alert_internal_bug()


class ErrFileMsg:
    @PreCheck(
        (lambda v: CheckRules.check_type_allow_none(v, str, PosixPath)), "filename"
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, int)), "line_no")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, int)), "col_no")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, int)), "end_col_no")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "src_code")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, int)), "indent_count")
    def __init__(
        self,
        filename: typing.Union[str, PosixPath] = None,
        line_no: int = None,
        col_no: int = None,
        end_col_no: int = None,
        src_code: str = None,
        arg_msg: str = DEFAULT_MSG_2,
        indent_count: int = 0,
        err_level: ErrLevel = ErrLevel.SERIOUS,
    ):
        self.filename = filename
        self.line_no = line_no
        self.col_no = col_no
        self.end_col_no = end_col_no
        self.src_code = src_code
        self.err_level = err_level if err_level else ErrLevel.SERIOUS
        self.arg_msg = arg_msg if arg_msg else DEFAULT_MSG_2
        self.indent_count = indent_count


@PreCheck((lambda v: CheckRules.check_type_not_none(v, ErrFileMsg)), "file_msg")
@PostCheck(lambda v: CheckRules.check_type_not_none(v, str))
def get_src_code(file_msg: ErrFileMsg) -> str:
    """Produce formatted error hint messages"""
    src_line = ""
    if file_msg.filename:
        source_line: str = ""
        if file_msg.line_no:
            if file_msg.src_code:
                lines = file_msg.src_code.split("\n")
            else:
                lines = (
                    open(file_msg.filename, "r", encoding="utf8").read().split("\n")
                    if os.path.exists(file_msg.filename)
                    else FileCache.get(file_msg.filename).split("\n")
                )

            if 0 < file_msg.line_no <= len(lines):
                source_line = lines[file_msg.line_no - 1]
            return source_line
    elif file_msg.src_code and file_msg.line_no:
        lines = file_msg.src_code.split("\n")
        if 0 < file_msg.line_no <= len(lines):
            src_line = lines[file_msg.line_no - 1]
        return src_line
    return ""


@PreCheck((lambda v: CheckRules.check_type_not_none(v, ErrFileMsg)), "file_msg")
@PostCheck(lambda v: CheckRules.check_type_not_none(v, str))
def get_hint_msg(
    file_msg: ErrFileMsg, msg_tem: KCLErrMsgTemplate = KCLErrMsgTemplateDefault()
) -> str:
    if msg_tem is None:
        msg_tem = KCLErrMsgTemplateDefault()
    msg_tem.clean_color()
    if msg_tem.fmt == ErrMsgFmt.COLOR_TXT:
        msg_tem.color_txt_err_file_msg(file_msg.err_level)
    result = ""
    arg = " -> " + file_msg.arg_msg if file_msg.arg_msg else ""
    mark = MARK_SYMBOL[file_msg.err_level]
    if file_msg.filename:
        if file_msg.line_no:
            src_code = get_src_code(file_msg)
            if file_msg.col_no and file_msg.end_col_no:
                result = (
                    msg_tem.get_hint_msg_detail(
                        file_msg.filename,
                        file_msg.line_no,
                        file_msg.col_no,
                        file_msg.end_col_no,
                        src_code,
                        indent_count=file_msg.indent_count,
                        mark=mark,
                    )
                    + arg
                )
            elif file_msg.col_no and not file_msg.end_col_no:
                result = (
                    msg_tem.get_hint_msg_common(
                        file_msg.filename,
                        file_msg.line_no,
                        file_msg.col_no,
                        src_code,
                        indent_count=file_msg.indent_count,
                        mark=mark,
                    )
                    + arg
                )
            elif not file_msg.col_no and not file_msg.end_col_no:
                result = (
                    msg_tem.get_hint_msg_summary(
                        file_msg.filename,
                        file_msg.line_no,
                        src_code,
                        indent_count=file_msg.indent_count,
                    )
                    + arg
                )
        else:
            result = (
                msg_tem.get_hint_msg_tiny(
                    file_msg.filename, indent_count=file_msg.indent_count
                )
                + arg
            )

    return result


@PreCheck((lambda v: CheckRules.check_type_not_none(v, str)), "ewcode")
@PostCheck(lambda v: CheckRules.check_type_not_none(v, str))
def get_err_msg(
    ewcode: str, msg_tem: KCLErrMsgTemplate = KCLErrMsgTemplateDefault()
) -> str:
    if msg_tem is None:
        msg_tem = KCLErrMsgTemplateDefault()
    msg_tem.clean_color()
    err_type = KCL_ERR_MSG.get_err_type_msg_by_errid(ewcode)
    err_cate = KCL_ERR_MSG.get_err_cate_msg_by_errid(ewcode)
    ewcode_fmt = KCL_ERR_MSG.get_err_code_by_errid(ewcode)
    err_msg = KCL_ERR_MSG.get_err_msg_by_errid(ewcode)

    if msg_tem.fmt == ErrMsgFmt.COLOR_TXT:
        msg_tem.color_txt_err_msg(err_type, err_cate)
    return msg_tem.err_msg_template(err_type, err_cate, ewcode_fmt, err_msg)



# Copyright 2021 The KCL Authors. All rights reserved.


class TokenValue:
    SEMI = ";"
    COMMA = ","
    EQ = "="
    LPAREN = "("
    RPAREN = ")"
    LBRACK = "["
    RBRACK = "]"
    LBRACE = "{"
    RBRACE = "}"
    LANGLE_BRACK = "<"
    RANGLE_BRACK = ">"
    SYNTAX = "syntax"
    IMPORT = "import"
    MESSAGE = "message"
    ENUM = "enum"
    PACKAGE = "package"
    OPTION = "option"
    ONEOF = "oneof"
    MAP = "map"
    REPEATED = "repeated"
    SERVICE = "service"
    RPC = "rpc"
    RETURNS = "returns"
    TO = "to"
    RESERVED = "reserved"



# Copyright 2021 The KCL Authors. All rights reserved.

from .builder import BuildLitNodeFromString, BuildLitNodeFromValue, BuildNodeFromString
from .filter import Declaration, filter_declarations, filter_stmt
from .fix import (
    fix_set_parent_info,
    fix_qualified_identifier,
    fix_and_get_module_import_list,
    fix_test_schema_auto_relaxed,
)

__all__ = [
    "BuildLitNodeFromString",
    "BuildLitNodeFromValue",
    "BuildNodeFromString",
    "Declaration",
    "filter_declarations",
    "filter_stmt",
    "fix_set_parent_info",
    "fix_qualified_identifier",
    "fix_and_get_module_import_list",
    "fix_test_schema_auto_relaxed",
]



from dataclasses import dataclass
from typing import List, Tuple, Union, cast
from copy import deepcopy

import kclvm.kcl.error as kcl
import kclvm.kcl.ast as ast
import kclvm.api.object as objpkg
from kclvm.vm.code import Opcode
from kclvm.compiler.check.check_type import type_pack_and_check
from kclvm.compiler.extension.builtin.system_module import json as _json
from kclvm.compiler.extension.builtin.system_module import yaml as _yaml

from .union import union, resolve_schema_obj
from .common import plus, handle_subscript


BINARY_FUNCTIONS = {
    Opcode.BINARY_ADD: lambda x, y: plus(x, y),
    Opcode.BINARY_SUBTRACT: lambda x, y: x - y,
    Opcode.BINARY_MULTIPLY: lambda x, y: x * y,
    Opcode.BINARY_TRUE_DIVIDE: lambda x, y: x / y,
    Opcode.BINARY_FLOOR_DIVIDE: lambda x, y: x // y,
    Opcode.BINARY_LSHIFT: lambda x, y: x << y,
    Opcode.BINARY_RSHIFT: lambda x, y: x >> y,
    Opcode.BINARY_POWER: lambda x, y: x**y,
    Opcode.BINARY_SUBSCR: lambda x, y: handle_subscript(x, y),
    Opcode.BINARY_XOR: lambda x, y: x ^ y,
    Opcode.BINARY_AND: lambda x, y: x & y,
    Opcode.BINARY_OR: lambda x, y: union(x, y),
    Opcode.BINARY_MODULO: lambda x, y: x % y,
    Opcode.BINARY_LOGIC_AND: lambda x, y: x and y,
    Opcode.BINARY_LOGIC_OR: lambda x, y: x or y,
    Opcode.COMPARE_EQUAL_TO: lambda x, y: x == y,
    Opcode.COMPARE_NOT_EQUAL_TO: lambda x, y: x != y,
    Opcode.COMPARE_LESS_THAN: lambda x, y: x < y,
    Opcode.COMPARE_LESS_THAN_OR_EQUAL_TO: lambda x, y: x <= y,
    Opcode.COMPARE_GREATER_THAN: lambda x, y: x > y,
    Opcode.COMPARE_GREATER_THAN_OR_EQUAL_TO: lambda x, y: x >= y,
    Opcode.COMPARE_IS: lambda x, y: x is y,
    Opcode.COMPARE_IS_NOT: lambda x, y: x is not y,
    Opcode.COMPARE_IN: lambda x, y: x in y,
    Opcode.COMPARE_NOT_IN: lambda x, y: x not in y,
}

UNARY_FUNCTIONS = {
    Opcode.UNARY_INVERT: lambda x: ~x,
    Opcode.UNARY_NOT: lambda x: not x,
    Opcode.UNARY_POSITIVE: lambda x: +x,
    Opcode.UNARY_NEGATIVE: lambda x: -x,
}

INPLACE_FUNCTIONS = {
    Opcode.INPLACE_ADD: lambda x, y: plus(x, y),
    Opcode.INPLACE_SUBTRACT: lambda x, y: x - y,
    Opcode.INPLACE_MULTIPLY: lambda x, y: x * y,
    Opcode.INPLACE_TRUE_DIVIDE: lambda x, y: x / y,
    Opcode.INPLACE_FLOOR_DIVIDE: lambda x, y: x // y,
    Opcode.INPLACE_LSHIFT: lambda x, y: x << y,
    Opcode.INPLACE_RSHIFT: lambda x, y: x >> y,
    Opcode.INPLACE_POWER: lambda x, y: x**y,
    Opcode.INPLACE_XOR: lambda x, y: x ^ y,
    Opcode.INPLACE_AND: lambda x, y: x & y,
    Opcode.INPLACE_OR: lambda x, y: union(x, y),
    Opcode.INPLACE_MODULO: lambda x, y: x % y,
}


@dataclass
class Evaluator:
    """Evaluator is a class responsible for parsing
    KCL objects and performing interpretation to get results
    """

    # Binary

    def eval_binary_op(
        self,
        left: objpkg.KCLObject,
        right: objpkg.KCLObject,
        code: Union[int, Opcode],
        vm=None,
    ) -> objpkg.KCLObject:
        if not left or not right or not code:
            raise Exception(f"invalid binary opcode action {left}, {right} and {code}")
        func = BINARY_FUNCTIONS.get(code)
        if not func:
            raise Exception(f"invalid binary opcode {code}")
        if code == Opcode.BINARY_OR:
            result = union(
                deepcopy(left),
                right,
                or_mode=True,
                should_config_resolve=True,
                should_idempotent_check=True,
                vm=vm,
            )
        elif code == Opcode.BINARY_ADD or code == Opcode.BINARY_SUBSCR:
            result = func(left, right)
        else:
            result = objpkg.to_kcl_obj(
                func(objpkg.to_python_obj(left), objpkg.to_python_obj(right))
            )
        return result

    # Unary

    def eval_unary_op(
        self, obj: objpkg.KCLObject, code: Union[int, Opcode]
    ) -> objpkg.KCLObject:
        if not obj or not code:
            raise Exception(f"invalid binary opcode action {obj} and {code}")
        func = UNARY_FUNCTIONS.get(code)
        if not func:
            raise Exception(f"invalid unary opcode {code}")
        r = objpkg.to_kcl_obj(func(obj.value))
        return r

    # Inplace operator

    def eval_inplace_op(
        self,
        left: objpkg.KCLObject,
        right: objpkg.KCLObject,
        code: Union[int, Opcode],
        vm=None,
    ) -> objpkg.KCLObject:
        if not left or not right or not code:
            raise Exception(f"invalid inpalce opcode action {left}, {right} and {code}")
        func = INPLACE_FUNCTIONS.get(code)
        if not func:
            raise Exception(f"invalid inplace opcode {code}")
        if code == Opcode.INPLACE_OR:
            result = union(
                left,
                right,
                or_mode=True,
                should_config_resolve=True,
                should_idempotent_check=True,
                vm=vm,
            )
        elif code == Opcode.INPLACE_ADD:
            result = func(left, right)
        else:
            result = objpkg.to_kcl_obj(
                func(
                    objpkg.to_python_obj(left.value), objpkg.to_python_obj(right.value)
                )
            )
        return result

    # Compare

    def eval_compare_op(
        self,
        left: objpkg.KCLObject,
        right: objpkg.KCLObject,
        code: Union[int, Opcode],
        vm=None,
    ) -> objpkg.KCLObject:
        # Avoid the overhead of boxing and unboxing large objects.
        if isinstance(right, (objpkg.KCLDictObject, objpkg.KCLSchemaObject)):
            if code == Opcode.COMPARE_IN:
                return objpkg.to_kcl_obj(left in right)
            elif code == Opcode.COMPARE_NOT_IN:
                return objpkg.to_kcl_obj(left not in right)
            else:
                return self.eval_binary_op(left, right, code, vm=vm)
        else:
            return self.eval_binary_op(left, right, code, vm=vm)

    # Iterable

    def iter_next(self, obj: objpkg.KCLObject) -> objpkg.KCLObject:
        if obj.type() != objpkg.KCLObjectType.ITER:
            kcl.report_exception(
                err_type=kcl.ErrType.EvaluationError_TYPE,
                arg_msg="only iterable object has next function",
            )
        return objpkg.to_kcl_obj(obj.next())

    # Functions

    def call_vars_and_keywords(
        self, argc: int, vm
    ) -> Tuple[List[objpkg.KCLObject], List[objpkg.KWArg]]:
        n_args = int(argc & 0xFF)
        n_kwargs = int((argc >> 8) & 0xFF)
        p, q = len(vm.stack) - 2 * n_kwargs, len(vm.stack)
        args = []
        kwargs = []
        for i in range((q - p) // 2 - 1, -1, -1):
            v = vm.pop()
            kstr = vm.pop()
            kwargs.append(objpkg.KWArg(name=kstr, value=v))
        p, q = p - n_args, p
        for i in range(q - p - 1, -1, -1):
            arg = vm.pop()
            args.append(arg)
        return args[::-1], kwargs[::-1]

    def eval_call(self, code: Union[int, Opcode], argc: int, vm) -> objpkg.KCLObject:
        if code == Opcode.CALL_FUNCTION:
            # Whether schema attribute value eval finalize.
            finalize = vm.pop().value
            # Function callable without `*args` and `**kwargs`
            args, kwargs = self.call_vars_and_keywords(argc, vm)
            callable_obj = vm.pop()
            if isinstance(callable_obj, objpkg.KCLCompiledFunctionObject):
                vm.push_frame_using_callable(
                    callable_obj.pkgpath,
                    callable_obj,
                    (args if args else []),
                    kwargs,
                    args_len=len(args),
                )
                return objpkg.NONE_INSTANCE
            elif isinstance(callable_obj, objpkg.KCLFunctionObject):
                result_obj = callable_obj.call(args, kwargs, vm)
                vm.push(result_obj)
                return result_obj
            elif isinstance(callable_obj, objpkg.KCLSchemaTypeObject):
                schema_type_obj = cast(objpkg.KCLSchemaTypeObject, callable_obj)
                inst = schema_type_obj.new_instance(
                    {}, {}, args, kwargs, vm, finalize=finalize
                )
                vm.push(inst)
                return inst
            elif isinstance(
                callable_obj, (objpkg.KCLNoneObject, objpkg.KCLUndefinedObject)
            ):
                # Ignore the user None callable
                vm.push(callable_obj)
                return callable_obj
            else:
                kcl.report_exception(
                    err_type=kcl.ErrType.EvaluationError_TYPE,
                    arg_msg=f"'{callable_obj.type()}' object is not callable",
                )

    # Attribute and subscript

    def load_attr(self, obj: objpkg.KCLObject, attr: str) -> objpkg.KCLObject:
        """Get attribute value of a KCL object"""
        if obj.type() in [
            objpkg.KCLObjectType.DICT,
            objpkg.KCLObjectType.SCHEMA,
            objpkg.KCLObjectType.MODULE,
        ]:
            return obj.get(attr)
        elif obj.type() in [
            objpkg.KCLObjectType.STRING,
            objpkg.KCLObjectType.SCHEMA_TYPE,
        ]:
            return obj.get_member_method(attr)
        else:
            kcl.report_exception(
                err_type=kcl.ErrType.EvaluationError_TYPE,
                arg_msg=f"'{obj.type_str()}' object has no attribute '{attr}'",
            )

    def set_attr(
        self,
        obj: objpkg.KCLObject,
        item: objpkg.KCLObject,
        value: objpkg.KCLObject,
        vm=None,
    ) -> objpkg.KCLObject:
        """Set attribute value of KCL object"""
        if not isinstance(obj, (objpkg.KCLDictObject, objpkg.KCLSchemaObject)):
            kcl.report_exception(
                err_type=kcl.ErrType.EvaluationError_TYPE,
                arg_msg="only schema and dict object can be updated attribute",
            )
        obj.update_key_value(item, value)
        if isinstance(obj, objpkg.KCLSchemaObject):
            obj = resolve_schema_obj(obj, obj.config_keys | {item.value}, vm=vm)
        return obj

    def set_item(
        self, obj: objpkg.KCLObject, item: objpkg.KCLObject, value: objpkg.KCLObject
    ) -> objpkg.KCLObject:
        """Set subscript value of a KCL object"""
        obj.value[item.value] = value
        return obj

    # List

    def list_append(
        self, obj: objpkg.KCLObject, item: objpkg.KCLObject
    ) -> objpkg.KCLObject:
        """Append an item into list"""
        if obj.type() != objpkg.KCLObjectType.LIST:
            kcl.report_exception(
                err_type=kcl.ErrType.EvaluationError_TYPE,
                arg_msg="only list object can append value",
            )
        obj.append(item)
        return obj

    # Dict

    def dict_append(
        self, obj: objpkg.KCLObject, key: objpkg.KCLObject, value: objpkg.KCLObject
    ) -> objpkg.KCLObject:
        """Append an key-value pair into dict"""
        if obj.type() != objpkg.KCLObjectType.DICT:
            if obj.type() != objpkg.KCLObjectType.LIST:
                kcl.report_exception(
                    err_type=kcl.ErrType.EvaluationError_TYPE,
                    arg_msg="only dict object can append key-value pair",
                )
        obj.update_key_value(key, value)
        return obj

    # String Format value

    def format_value(
        self, obj: objpkg.KCLObject, format_spec: str
    ) -> objpkg.KCLStringObject:
        if not format_spec:
            value = "{}".format(objpkg.to_python_obj(obj))
            return objpkg.KCLStringObject(value=value)
        assert isinstance(format_spec, str)
        if format_spec.lower() == "#json":
            value = _json.KMANGLED_encode(objpkg.to_python_obj(obj))
            return objpkg.KCLStringObject(value=value)
        if format_spec.lower() == "#yaml":
            value = _yaml.KMANGLED_encode(objpkg.to_python_obj(obj))
            return objpkg.KCLStringObject(value=value)
        kcl.report_exception(
            err_type=kcl.ErrType.InvalidFormatSpec_TYPE,
            arg_msg=kcl.INVALID_FORMAT_SPEC_MSG.format(format_spec),
        )

    # Schema config operation expression

    def update_schema_attr(
        self,
        attr: str,
        schema_obj: objpkg.KCLSchemaObject,
        config_value: objpkg.KCLObject,
        conf_meta: dict,
        expected_types: List[str],
        operation=ast.ConfigEntryOperation.UNION,
        index=None,
        vm=None,
        filename=None,
        lineno=None,
        columnno=None,
    ):
        """Update schema attr value with config dict"""
        if index is None or index < 0:
            config_value_checked = type_pack_and_check(
                config_value,
                expected_types,
                vm=vm,
                filename=filename,
                lineno=lineno,
                columno=columnno,
                config_meta=conf_meta.get("$conf_meta"),
            )
        else:
            value = schema_obj.get(attr)
            if not isinstance(value, objpkg.KCLListObject):
                kcl.report_exception(
                    err_type=kcl.ErrType.EvaluationError_TYPE,
                    arg_msg="only list attribute can be inserted value",
                )
            config_value_checked = type_pack_and_check(
                objpkg.KCLListObject(items=[config_value]),
                expected_types,
                vm=vm,
                filename=filename,
                lineno=lineno,
                columno=columnno,
                config_meta=conf_meta.get("$conf_meta"),
            )
            config_value_checked = cast(
                objpkg.KCLListObject, config_value_checked
            ).items[0]
        SCHEMA_CONFIG_OP_MAPPING = {
            ast.ConfigEntryOperation.UNION: self.union_schema_attr,
            ast.ConfigEntryOperation.OVERRIDE: self.override_schema_attr,
            ast.ConfigEntryOperation.INSERT: self.insert_schema_attr,
        }
        func = SCHEMA_CONFIG_OP_MAPPING.get(operation)
        if not func:
            raise Exception(f"Invalid schema config object operation: {operation}")
        func(attr, schema_obj, config_value_checked, index, vm)

    def union_schema_attr(
        self,
        attr: str,
        schema_obj: objpkg.KCLSchemaObject,
        config_value: objpkg.KCLObject,
        index=None,
        vm=None,
    ):
        if index is None or index < 0:
            # TODO: modify `should_idempotent_check` to False after Konfig code update finish
            schema_obj.union_with({attr: config_value}, should_idempotent_check=False)
        else:
            # Union with list internal index value
            value = schema_obj.get(attr)
            value.items[index] = union(
                value.items[index], config_value, should_idempotent_check=False, vm=vm
            )
            schema_obj.union_with({attr: value}, should_idempotent_check=False)

    def override_schema_attr(
        self,
        attr: str,
        schema_obj: objpkg.KCLSchemaObject,
        config_value: objpkg.KCLObject,
        index=None,
        vm=None,
    ):
        if index is None or index < 0:
            schema_obj.update_key_value(attr, config_value)
        else:
            schema_obj.list_key_override(attr, config_value, index)

    def insert_schema_attr(
        self,
        attr: str,
        schema_obj: objpkg.KCLSchemaObject,
        config_value: objpkg.KCLObject,
        index=None,
        vm=None,
    ):
        if index is not None and index >= 0:
            config_value = objpkg.KCLListObject(items=[config_value])
        schema_obj.insert_with(attr, config_value, index)



"""
KCLLinter class controls all inspection processes of lint: loading config, checking and generating reports.

The workflow of KCLLinter is as follows:
1. Load config.
2. Find all KCL files under the 'path' from CLI arguments, and parse them to ast.Program.
3. Register checker and reporter according to config
4. Distribute ast to each checker for checking, and generate Message，which represents the result of check.
5. Linter collects Messages from all checkers.
6. Distribute Message to each reporter as output
┌─────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                   KCLLinter                                                                 │
│                                                                                                             │
│      ┌───────────┐                  ┌─────────────────────────────────────────────────────────────────┐     │
│      │  KCL file │                  │                             Checker                             │     │
│      └───────────┘                  │  ┌─────────────────┐  ┌─────────────────┐  ┌─────────────────┐  │     │
│            ↓                        │  │  importChecker  │  │  schemaChecker  │  │       ...       │  │     │
│      ┌───────────┐                  │  │  ┌───────────┐  │  │  ┌───────────┐  │  │  ┌───────────┐  │  │     │
│      │  ast.Prog │       →          │  │  │  Message  │  │  │  │  Message  │  │  │  │  Message  │  │  │     │
│      └───────────┘                  │  │  └───────────┘  │  │  └───────────┘  │  │  └───────────┘  │  │     │
│                                     │  │  ┌───────────┐  │  │  ┌───────────┐  │  │  ┌───────────┐  │  │     │
│                                     │  │  │  Message  │  │  │  │  Message  │  │  │  │  Message  │  │  │     │
│                                     │  │  └───────────┘  │  │  └───────────┘  │  │  └───────────┘  │  │     │
│      ┌──────────────────────┐       │  │  ┌───────────┐  │  │  ┌───────────┐  │  │  ┌───────────┐  │  │     │
│      │      Config          │       │  │  │    ...    │  │  │  │    ...    │  │  │  │    ...    │  │  │     │
│      │                      │       │  │  └───────────┘  │  │  └───────────┘  │  │  └───────────┘  │  │     │
│      │   1 config           │       │  └─────────────────┘  └─────────────────┘  └─────────────────┘  │     │
│      │   2 .kcllint         │       └─────────────────────────────────────────────────────────────────┘     │
│      │   3 default_config   │                                                                               │
│      │                      │                                        ↓                                      │
│      │                      │       msgs_map -> MessageID: count                                            │
│      └──────────────────────┘       msgs ->    ┌────────────────────────────────────────────────────┐       │
│                                                │  ┌───────────┐  ┌───────────┐  ┌───────────┐       │       │
│                                                │  │  Message  │  │  Message  │  │  Message  │       │       │
│                                                │  └───────────┘  └───────────┘  └───────────┘       │       │
│                                                └────────────────────────────────────────────────────┘       │
│                                                                                                             │
│                                                                      ↓                                      │
│                                     ┌─────────────────────────────────────────────────────────────────┐     │
│                                     │                              Reporter                           │     │
│                                     │  ┌───────────┐  ┌───────────┐  ┌───────────┐  ┌───────────┐     │     │
│                                     │  │  stdout   │  │   sarif   │  │   file    │  │   ...     │     │     │
│                                     │  └───────────┘  └───────────┘  └───────────┘  └───────────┘     │     │
│                                     └─────────────────────────────────────────────────────────────────┘     │
│                                                                                                             │
│                                                                                                             │
│                                                                                                             │
└─────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
"""
import os
import glob
import ruamel.yaml as yaml
from typing import Dict, Any, Optional, List

import kclvm.compiler.parser.parser as parser
from kclvm.tools.lint.checkers import *
import kclvm.tools.lint.reporters.base_reporter as base_reporter
import kclvm.tools.lint.reporters.stdout_reporter as stdout_reporter
import kclvm.tools.lint.reporters.file_reporter as file_reporter
import kclvm.tools.lint.reporters.sarif_reporter as sarif_reporter
import kclvm.kcl.error.kcl_error as error
import kclvm.tools.lint.message.message as message
import kclvm.tools.lint.lint.exceptions as exceptions
import kclvm.kcl.info as kcl_info
import kclvm.kcl.ast as ast

LINT_CONFIG_SUFFIX = ".kcllint"
DEFAULT_CONFIG = {
    "check_list": ["import", "misc", "basic"],
    "ignore": [],
    "max_line_length": 200,
    "output": ["stdout"],
    "output_path": None,
    "module_naming_style": "ANY",
    "package_naming_style": "ANY",
    "schema_naming_style": "PascalCase",
    "mixin_naming_style": "PascalCase",
    "protocol_naming_style": "PascalCase",
    "argument_naming_style": "camelCase",
    "variable_naming_style": "ANY",
    "schema_attribute_naming_style": "ANY",
    "module_rgx": None,
    "package_rgx": None,
    "schema_rgx": None,
    "mixin_rgx": None,
    "protocol_rgx": None,
    "argument_rgx": None,
    "variable_rgx": None,
    "schema_attribute_rgx": None,
    "bad_names": ["foo", "bar", "baz", "toto", "tata", "tutu", "I", "l", "O"],
}

MSGS = {"E0999": ("Parse failed:%s.", "Parse failed.", "Parse failed:'{0}'.")}
PARSE_FAILED_MSG_ID = "E0999"


class CheckerFactory:
    @staticmethod
    def get_checker(checker: str, linter=None) -> Optional[base_checker.BaseChecker]:
        if checker == "import":
            return imports.ImportsChecker(linter)
        elif checker == "misc":
            return misc.MiscChecker(linter)
        elif checker == "basic":
            return basic.BasicChecker(linter)
        else:
            raise exceptions.InvalidCheckerError(checker)


class ReporterFactory:
    @staticmethod
    def get_reporter(
        reporter: str, linter=None
    ) -> Optional[base_reporter.BaseReporter]:
        if reporter == "stdout":
            return stdout_reporter.STDOUTReporter(linter)
        elif reporter == "file":
            return file_reporter.FileReporter(linter)
        elif reporter == "sarif":
            return sarif_reporter.SARIFReporter(linter)
        else:
            raise exceptions.InvalidReporterError(reporter)


class LinterConfig:
    def __init__(self):
        self.check_list = ["import", "misc", "basic"]
        self.ignore = []
        self.max_line_length = 200
        self.output = ["stdout"]
        self.output_path = None
        self.module_naming_style = "ANY"
        self.package_naming_style = "ANY"
        self.schema_naming_style = "PascalCase"
        self.mixin_naming_style = "PascalCase"
        self.protocol_naming_style = "PascalCase"
        self.argument_naming_style = "camelCase"
        self.variable_naming_style = "ANY"
        self.schema_attribute_naming_style = "ANY"
        self.module_rgx = None
        self.package_rgx = None
        self.schema_rgx = None
        self.mixin_rgx = None
        self.protocol_rgx = None
        self.argument_rgx = None
        self.variable_rgx = None
        self.schema_attribute_rgx = None
        self.bad_names = ["foo", "bar", "baz", "toto", "tata", "tutu", "I", "l", "O"]

    def update(self, config: {}):
        for k, v in config.items():
            if hasattr(self, k):
                self.__setattr__(k, v)


class KCLLinter:
    def __init__(
        self,
        *path: str,
        config: Dict[str, Any] = None,
        k_code_list: List[str] = None,
    ) -> None:
        self.path = path or []
        self.k_code_list = k_code_list or []
        self.file_list = []
        self.programs_list = []
        self.checkers = []
        self.reporters = []
        self.config = LinterConfig()
        self.msgs = []
        self.MSGS = MSGS
        self.msgs_map = {}

        path_list = [x for x in path]
        for i, s in enumerate(path_list):
            s = os.path.abspath(s)
            if os.path.isfile(s):
                self.file_list.append(s)
            elif os.path.isdir(s):
                self.file_list += glob.glob(
                    os.path.join(s, "**", kcl_info.KCL_FILE_PATTERN),
                    recursive=True,
                )
            else:
                raise FileNotFoundError(s)

        self._load_config(config)

    def reset(self):
        self.reporters.clear()
        self.checkers.clear()
        self.MSGS = MSGS
        self.msgs.clear()
        self.msgs_map = {}

    def run(self) -> None:
        self.reset()
        self._register_checkers(self.config.check_list)
        self._register_reporters(self.config.output)
        self._get_programs(self.file_list, self.k_code_list)
        self._check(self.programs_list, self.checkers, self.k_code_list)
        self._display()

    def check(self) -> None:
        self.reset()
        self._register_checkers(self.config.check_list)
        self._register_reporters(self.config.output)
        self._get_programs(self.file_list, self.k_code_list)
        self._check(self.programs_list, self.checkers, self.k_code_list)

    def _load_config(self, config) -> None:
        for s in self.path:
            if os.path.isfile(s):
                kcllint_path = os.path.join(
                    os.path.abspath(os.path.dirname(s)), LINT_CONFIG_SUFFIX
                )
            elif os.path.isdir(s):
                kcllint_path = os.path.join(os.path.abspath(s), LINT_CONFIG_SUFFIX)
            if os.path.isfile(kcllint_path):
                with open(kcllint_path, "r", encoding="utf-8") as f:
                    kcllint_config = f.read()
                    self.config.update(yaml.safe_load(kcllint_config))
                break
        if config:
            self.config.update(config)

    def _register_checker(self, checker: base_checker.BaseChecker) -> None:
        self.checkers.append(checker)
        if hasattr(checker, "MSGS"):
            self.MSGS.update(checker.MSGS)

    def _register_checkers(self, checkers: List[str]) -> None:
        factory = CheckerFactory()
        for s in checkers:
            self._register_checker(factory.get_checker(s, self))

    def _register_reporters(self, reporters: List[str]) -> None:
        if not reporters or len(reporters) == 0:
            raise exceptions.EmptyReporterError
        if "file" in reporters or "sarif" in reporters:
            assert self.config.output_path, "Without ouput file path"
        factory = ReporterFactory()
        self.reporters = [factory.get_reporter(s, self) for s in reporters]

    def _get_programs(
        self, file_list: List[str], k_code_list: List[str]
    ) -> List[ast.Program]:
        for i, file in enumerate(file_list):
            _code = k_code_list[i] if (i < len(k_code_list)) else None
            _k_code_list = [_code] if _code else None
            try:
                prog = parser.LoadProgram(file, k_code_list=_k_code_list)
                self.programs_list.append(prog)
            except error.KCLException as err:
                if not _code:
                    with open(err.filename) as f:
                        _code = f.read()
                source_line = _code.split("\n")[err.lineno - 1]
                msg = message.Message(
                    PARSE_FAILED_MSG_ID,
                    err.filename,
                    MSGS[PARSE_FAILED_MSG_ID][0] % err.name,
                    source_line,
                    (err.lineno, err.colno),
                    [err.name],
                )
                if (msg not in self.msgs) and (
                    PARSE_FAILED_MSG_ID not in self.config.ignore
                ):
                    self.msgs.append(msg)
                    self.msgs_map[PARSE_FAILED_MSG_ID] = (
                        self.msgs_map.setdefault(PARSE_FAILED_MSG_ID, 0) + 1
                    )

    def _check(
        self,
        progs: List[ast.Program],
        checkers: List[base_checker.BaseChecker],
        k_code_list: List[str],
    ):
        for i, prog in enumerate(progs):
            if i < len(k_code_list):
                _code = k_code_list[i]
            else:
                module = prog.pkgs["__main__"][0]
                with open(module.filename) as f:
                    _code = f.read()
            for checker in checkers:
                checker.check(prog, _code)
                # collect msgs to linter
                for msg in checker.msgs:
                    if msg.msg_id in self.config.ignore:
                        continue
                    if msg not in self.msgs:
                        self.msgs.append(msg)
                        self.msgs_map[msg.msg_id] = (
                            self.msgs_map.setdefault(msg.msg_id, 0) + 1
                        )

    def _display(self) -> None:
        for reporter in self.reporters:
            reporter.display()


def kcl_lint(*path: str, config: Dict[str, Any] = None) -> List[message.Message]:
    """
    Check kcl files or all kcl files in dirs
    :param path: str, path of a kcl file or dir
    :param config: Dict[str, Any], config of lint
    :return: List[Message] result of lint check
    """
    lint = KCLLinter(*path, config=config)
    lint.check()
    return lint.msgs


def kcl_lint_code(
    *path: str,
    k_code_list: List[str],
    config: Dict[str, Any] = None,
) -> List[message.Message]:
    """
    Check individual code of string type or some code in kcl file
    if pararm:file not None, file must be a path of kcl file and code should be part of the file,
    .e.g select some code for checking in ide
    :param k_code_list: code of string type or some code in kcl file
    :param config: Dict[str, Any], config of lint
    :param path: path of kcl file
    :return: List[Message] result of lint check
    """
    assert len(k_code_list) > 0
    lint = KCLLinter(*path, config=config, k_code_list=k_code_list)
    lint.check()
    return lint.msgs



# Copyright 2021 The KCL Authors. All rights reserved.

import sys
import enum
from abc import ABCMeta, abstractmethod
from typing import Dict, Type

import kclvm.kcl.error as kcl_error

DECORATOR_TARGET_ERR_NAME_MSG = ": Decorator target name cannot be empty"


class DecoratorTargetType(enum.Enum):
    """
    Marked annotated position by the decorator
    """

    SCHEMA_TYPE = 0
    ATTRIBUTE = 1


class Decorator(metaclass=ABCMeta):
    """
    An abstract decorator.

    This abc is used to run actions as a wrapper of key-value pair in kcl schema.

    A concrete decorator should inherit this class and impel the run method to handle key-value.

    :class:`~Deprecated` class is a sample of decorator which is used to check deprecation of key on config.

    you can use your ``Deprecated`` class like this in your source code as follows:

    .. code-block:: python

        schema Person:
            @deprecated(version="1.16", reason="use firstName and lastName instead", strict=True)
            name : str

    .. code-block:: python

        @deprecated(version="1.16", reason="use firstName and lastName instead", strict=True)
        schema Person:
            name : str

    """

    def __init__(self, name: str, target: DecoratorTargetType, *args, **kwargs):
        self.name = name
        try:
            self.target = (
                target
                if isinstance(target, DecoratorTargetType)
                else DecoratorTargetType(target)
            )
        except ValueError:
            msg = (
                kcl_error.INVALID_DECORATOR_TARGET_MSG.format(target)
                if target
                else kcl_error.INVALID_DECORATOR_TARGET_MSG.format("")
            )
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.InvalidDecoratorTarget_TYPE,
                file_msgs=[
                    kcl_error.ErrFileMsg(
                        filename=kwargs.get("filename"),
                        line_no=kwargs.get("lineno"),
                        col_no=kwargs.get("columnno"),
                    )
                ],
                arg_msg=msg,
            )
        self.filename = kwargs.get("filename")
        self.lineno = kwargs.get("lineno")
        self.columnno = kwargs.get("columnno")

    @abstractmethod
    def run(self, key: str, value, *args, **kwargs):
        """
        Decorate on a key-value pair in runtime.

        :param key: Key to Decorated
        :param value: Value to Decorated
        :param args: Decorator run positional args
        :param kwargs: Decorator run keyword args
        :return:
        """
        pass


class DecoratorFactory:
    """
    A decorator factory used to get decorator object.
    """

    def __init__(self):
        self._decorators: Dict[str, Type[Decorator]] = {}

    def register(self, name: str, decorator: Type[Decorator]):
        """
        Register a decorator with a unique name.

        :param name: Name of the decorator
        :param decorator: The decorator to be registered
        :return: None
        """
        self._decorators[name] = decorator

    def get(self, name: str, target: DecoratorTargetType, *args, **kwargs):
        """
        Get and return a decorator object.
        An UnKnownDecorator will be thrown if no decorator found.

        :param name: Name of the decorator
        :param target: Target of the decorator e.g., schema and attribute
        :param args: Decorator meta info positional args
        :param kwargs: Decorator meta info keyword args
        :return: A decorator object
        """
        decorator = self._decorators.get(name)
        if not decorator:
            filename = kwargs.get("filename")
            lineno = kwargs.get("lineno")
            columnno = kwargs.get("columnno")
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.UnKnownDecorator_TYPE,
                file_msgs=[
                    kcl_error.ErrFileMsg(
                        filename=filename, line_no=lineno, col_no=columnno
                    )
                ],
                arg_msg=kcl_error.UNKNOWN_DECORATOR_MSG.format(name)
                if name
                else kcl_error.UNKNOWN_DECORATOR_MSG.format(""),
            )
        return decorator(name, target, *args, **kwargs)


class Deprecated(Decorator):
    """This decorator is used to get the deprecation message according to the wrapped key-value pair.

    Examples
    --------
        @deprecated(version="v1.16", reason="The age attribute was deprecated", strict=True)
        schema Person:
            name: str
            age: int
    """

    NAME = "deprecated"

    def __init__(self, name, target, *args, **kwargs):
        """
        Construct a deprecated decorator

        :param args: Deprecated decorator build positional args
        :param kwargs: Deprecated decorator build keyword args
        """
        super().__init__(self.NAME, target, *args, **kwargs)
        self.version = kwargs.get("version", "")
        self.reason = kwargs.get("reason", "")
        self.strict = kwargs.get("strict", True)

    def run(self, key: str, value, *args, **kwargs):
        """
        Build and report deprecation message.

        A KCL runtime error will be thrown if self.strict is True, otherwise it just print warning message.

        :param key: Key to Deprecated decorated
        :param value: Value to Deprecated decorated
        :param args: Deprecated decorator run positional args
        :param kwargs: Deprecated decorator run positional args
        :return should_change_value: Mark the value of the deprecated schema attribute should be modified
        """
        filename = self.filename
        lineno = self.lineno
        columnno = self.columnno
        if not key:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.NameError_TYPE,
                file_msgs=[
                    kcl_error.ErrFileMsg(
                        filename=filename, line_no=lineno, col_no=columnno
                    )
                ],
                arg_msg=kcl_error.NAME_ERROR_MSG.format(DECORATOR_TARGET_ERR_NAME_MSG),
            )
        # Mark the value of the deprecated schema attribute should be modified
        should_change_value = False
        # Error or warning message
        msg = ""
        # Append a version info into message
        if self.version:
            msg += "since version {}".format(self.version)
        # Append a reason info into message
        if self.reason:
            msg += ", " + self.reason
        if self.strict:
            if self.target == DecoratorTargetType.SCHEMA_TYPE or (
                self.target == DecoratorTargetType.ATTRIBUTE and value
            ):
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.Deprecated_TYPE,
                    file_msgs=[
                        kcl_error.ErrFileMsg(
                            filename=filename, line_no=lineno, col_no=columnno
                        )
                    ],
                    arg_msg=kcl_error.DEPRECATED_WARNING_MSG.format(key, msg),
                )
            should_change_value = True
        else:
            # If it is a modified schema attribute, ignore the assignment without reporting an error.
            kcl_error.print_kcl_warning_message(
                kcl_error.get_exception(
                    err_type=kcl_error.ErrType.Deprecated_Warning_TYPE,
                    file_msgs=[
                        kcl_error.ErrFileMsg(
                            filename=filename, line_no=lineno, col_no=columnno
                        )
                    ],
                    arg_msg=kcl_error.DEPRECATED_WARNING_MSG.format(key, msg),
                ),
                file=sys.stderr,
            )
            should_change_value = False
        return should_change_value


class Info(Decorator):
    """Info decorator is used to mark some compile-time information for external API queries

    Examples
    --------
        @info(message="User message")
        schema Person:
            name: str
            age: int
    """

    NAME = "info"

    def __init__(self, name, target, *args, **kwargs):
        """Construct a Info decorator"""
        super().__init__(self.NAME, target, *args, **kwargs)

    def run(self, key: str, value, *args, **kwargs):
        """Nothing to do on Info decorator"""
        pass


decorator_factory = DecoratorFactory()
decorator_factory.register(Deprecated.NAME, Deprecated)
decorator_factory.register(Info.NAME, Info)



"""
Base id of standard checkers (used in msg and report ids):
01: basic
02: Schema
03: format
04: import
05: misc
06: variables
07:
08：
09: syntax
10-50: not yet used: reserved for future internal checkers.
51-99: perhaps used: reserved for external checkers
"""

__all__ = ["base_checker", "imports", "misc", "basic"]



import os
from typing import Union

import kclvm.kcl.info as kcl_info
import kclvm.tools.lint.message.message as message
import kclvm.compiler.vfs as vfs
import kclvm.kcl.ast as ast
import kclvm.compiler.extension.plugin.plugin_model as plugin
import kclvm.compiler.extension.builtin.builtin as builtin
import kclvm.tools.lint.lint.utils as utils
from kclvm.tools.lint.checkers.base_checker import BaseChecker
import re

MSGS = {
    "E0401": ("Unable to import %s.", "Unable to import.", "Unable to import '{0}'."),
    "W0404": (
        "%s is reimported multiple times.",
        "Module reimported.",
        "'{0}' is reimported multiple times.",
    ),
    "E0406": (
        "Module import itself.",
        "Module import itself.",
        "Module import itself.",
    ),
    "W0411": (
        "%s imported but unused.",
        "Module imported but unused.",
        "'{0}' imported but unused.",
    ),
    "E0413": (
        "Import %s should be placed at the top of the module.",
        "ImportStmt is not at the top of the file.",
        "Import '{0}' should be placed at the top of the module.",
    ),
}

IMPORT_POSITION_CHECK_LIST = [
    "AssignStmt",
    "AugAssignStmt",
    "AssertStmt",
    "IfStmt",
    "TypeAliasStmt",
    "SchemaStmt",
    "RuleStmt",
]


class ImportsChecker(BaseChecker):
    def __init__(self, linter=None) -> None:
        super().__init__(linter)
        self.name = "ImportCheck"
        self.MSGS = MSGS
        self.prog = None
        self.module = None
        self.code = None
        self.root: str = None
        # for reimport check
        self.has_imported_modules = None
        # for unused import check
        self.import_names_map = None
        # for import position check
        self.import_position_check = True

    def reset(self) -> None:
        self.msgs.clear()
        self.module = None
        self.code = None
        self.root = None
        if self.import_names_map:
            self.import_names_map.clear()
        if self.has_imported_modules:
            self.has_imported_modules.clear()
        self.import_position_check = True

    def get_module(self, prog: ast.Program, code: str) -> None:
        assert isinstance(prog, ast.Program)
        assert code is not None
        self.prog = prog
        self.module = prog.pkgs["__main__"][0]
        self.code = code
        self.root = prog.root
        self.has_imported_modules = []
        self.import_names_map = {}

    def check(self, prog: ast.Program, code: str) -> None:
        assert isinstance(prog, ast.Program)
        assert code is not None
        self.reset()
        self.get_module(prog, code)
        self.walk(self.module)
        for k, v in self.import_names_map.items():
            self.msgs.append(
                message.Message(
                    "W0411",
                    self.module.filename,
                    MSGS["W0411"][0] % k,
                    utils.get_source_code(self.module.filename, v.line, self.code),
                    (v.line, v.column),
                    [k],
                )
            )

    def check_import_file_exist(self, t: ast.ImportStmt, abs_path: str) -> bool:
        assert isinstance(t, ast.ImportStmt)
        if os.path.isdir(abs_path) or os.path.isfile(
            abs_path + kcl_info.KCL_FILE_SUFFIX
        ):
            return True
        else:
            self.msgs.append(
                message.Message(
                    "E0401",
                    self.module.filename,
                    MSGS["E0401"][0] % t.path,
                    utils.get_source_code(self.module.filename, t.line, self.code),
                    (t.line, t.column),
                    [t.path],
                )
            )
            return False

    def check_import_position(self, t: ast.AST) -> None:
        assert isinstance(t, ast.AST)
        if self.import_position_check:
            if t._ast_type in IMPORT_POSITION_CHECK_LIST:
                self.import_position_check = False
        else:
            if isinstance(t, ast.ImportStmt):
                self.msgs.append(
                    message.Message(
                        "E0413",
                        self.module.filename,
                        MSGS["E0413"][0] % t.pkg_name,
                        utils.get_source_code(self.module.filename, t.line, self.code),
                        (t.line, t.column),
                        [t.pkg_name],
                    )
                )

    def check_unused_import(self, t: Union[ast.Identifier, str]) -> None:
        if isinstance(t, ast.Identifier):
            if t.get_first_name() in self.import_names_map.keys():
                self.import_names_map.pop(t.get_first_name())
        else:
            # SchemaAttr.types, A|B, [A|B], {A|B:C}
            type_list = re.split(r"[|:\[\]\{\}]", t)
            for type in type_list:
                names = type.split(".")
                first_name = names[0]
                if first_name in self.import_names_map.keys():
                    self.import_names_map.pop(first_name)

    def check_import_itself(self, t: ast.ImportStmt, abs_path: str) -> None:
        assert isinstance(t, ast.ImportStmt)
        if os.path.isdir(abs_path):
            return
        abs_path += kcl_info.KCL_FILE_SUFFIX
        # normpath: a/./b -> a/b
        if abs_path == str(os.path.normpath(self.module.filename)):
            self.msgs.append(
                message.Message(
                    "E0406",
                    self.module.filename,
                    MSGS["E0406"][0],
                    utils.get_source_code(self.module.filename, t.line, self.code),
                    (t.line, t.column),
                    [],
                )
            )

    def check_reimport(self, t: ast.ImportStmt, abs_path: str) -> None:
        assert isinstance(t, ast.ImportStmt)
        if abs_path in self.has_imported_modules:
            self.msgs.append(
                message.Message(
                    "W0404",
                    self.module.filename,
                    MSGS["W0404"][0] % t.pkg_name,
                    utils.get_source_code(self.module.filename, t.line, self.code),
                    (t.line, t.column),
                    [t.pkg_name],
                )
            )
        else:
            self.import_names_map[t.pkg_name] = t
            self.has_imported_modules.append(abs_path)

    def generic_walk(self, t: ast.AST) -> None:
        """Called if no explicit walker function exists for a node."""
        self.check_import_position(t)
        for field, value in ast.iter_fields(t):
            if isinstance(value, list):
                for v in value:
                    # IfStmt.elif_body: List[List[Stmt]]
                    if isinstance(v, list):
                        for v1 in v:
                            self.walk(v1)
                    if isinstance(v, ast.AST):
                        self.walk(v)
            elif isinstance(value, ast.AST):
                self.walk(value)

    def walk_Identifier(self, t: ast.Identifier) -> None:
        assert isinstance(t, ast.Identifier)
        self.check_unused_import(t)

    def walk_SchemaAttr(self, t: ast.SchemaAttr) -> None:
        assert isinstance(t, ast.SchemaAttr)
        self.check_unused_import(t.type_str)
        self.generic_walk(t)

    def walk_ImportStmt(self, t: ast.ImportStmt) -> None:
        assert isinstance(t, ast.ImportStmt)
        self.check_import_position(t)
        if (
            t.path.startswith(plugin.PLUGIN_MODULE_NAME)
            or t.name in builtin.STANDARD_SYSTEM_MODULES
        ):
            self.check_reimport(t, t.path)
            return
        fix_path = vfs.FixImportPath(self.root, self.module.filename, t.path).replace(
            ".", os.sep
        )
        abs_path = os.path.join(self.root, fix_path)
        if self.check_import_file_exist(t, abs_path):
            self.check_import_itself(t, abs_path)
            self.check_reimport(t, abs_path)



# Copyright 2020 The KCL Authors. All rights reserved.

from .parser import ParseMode, ParseFile, ParseExpr, LoadProgram


__all__ = [
    "ParseMode",
    "ParseFile",
    "ParseExpr",
    "LoadProgram",
]



# Copyright 2021 The KCL Authors. All rights reserved.

import typing
import pathlib
import os
import sys

import toml
import google.protobuf.json_format as json_format

import kclvm.config
import kclvm.kcl.error as kcl_error


def GetPkgRoot(
    k_file_path: str, should_current_file_work_dir: bool = True
) -> typing.Optional[str]:
    """Search kcl.mod filepath with the KCL file path"""
    if not k_file_path:
        return None

    # search by kcl.mod file
    module_path = pathlib.Path(os.path.abspath(k_file_path))
    root = module_path.root
    while module_path:
        if module_path == module_path.parent or str(module_path) == root:
            break

        kcl_mod_path = module_path.joinpath("kcl.mod")
        if kcl_mod_path.exists() and kcl_mod_path.is_file():
            return str(module_path)

        module_path = module_path.parent

    if should_current_file_work_dir and k_file_path.endswith(".k"):
        return os.path.dirname(k_file_path)

    return None


def MustGetPkgRoot(file_paths: typing.List[str]) -> typing.Optional[str]:
    """Search kcl.mod filepath with the KCL file paths,
    when found multiple kcl.mod paths, raise a compile error.
    """
    # Get kcl.mod paths from input file paths and remove empty path using the filter function.
    paths = set(
        filter(
            None,
            [
                GetPkgRoot(file_path, should_current_file_work_dir=False)
                for file_path in file_paths or []
            ],
        )
    )
    # Not found kcl.mod.
    if not paths:
        return None
    # Find one kcl.mod.
    if len(paths) == 1:
        return list(paths)[0]
    # Find multiple kcl.mod, raise an error.
    kcl_error.report_exception(
        err_type=kcl_error.ErrType.CompileError_TYPE,
        arg_msg=f"conflict kcl.mod file paths: {paths}",
    )


def LoadModFile(root: str) -> kclvm.config.KclModFile:
    k_mod_file_path = f"{root}{os.sep}kcl.mod"

    if not os.path.exists(k_mod_file_path):
        mod_file = kclvm.config.KclModFile(root=root)
        return mod_file

    d = toml.load(k_mod_file_path)
    mod_file = kclvm.config.KclModFile(root=root)
    json_format.ParseDict(d, mod_file, ignore_unknown_fields=True)
    mod_file.root = root
    return mod_file


if __name__ == "__main__":
    if len(sys.argv) < 2 or (sys.argv[1] == "-h" or sys.argv[1] == "-help"):
        print("usage: python3 ./this_py_file <kcl.mod root>")
        sys.exit(0)

    f = LoadModFile(sys.argv[1])
    print(f)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import pathlib

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from pkg_resources import require
import setuptools

# Steps:
# 1. modify kclvm version in setup.py
# 2. run `python3 setup.py sdist` to build package
# 3. run `python3 -m twine upload dist/kclvm-*.tar.gz` to upload package to PyPI
# 4. input username and password of PyPI

install_requires = []
require_path = pathlib.Path(__file__).parent.joinpath("kclvm/scripts/requirements.txt")
with open(require_path) as f:
    requires = f.read().split("\n")
    for require in requires:
        install_requires.append(require)

setup(
    name="kclvm",
    author="KCL Authors",
    version="0.9.0-beta.1",
    license="Apache License 2.0",
    python_requires=">=3.7",
    description="KCL Python SDK",
    long_description="""A constraint-based record & functional language mainly used in configuration and policy scenarios.""",
    author_email="",
    data_files=[
        "kclvm/tools/docs/templates/md.mako",
        "kclvm/scripts/requirements.txt",
        "kclvm/api/version/checksum.txt",
        "kclvm/kcl/grammar/kcl.lark",
        "kclvm/tools/docs/model.proto",
        "kclvm/encoding/protobuf/kcl.proto",
        "kclvm/spec/gpyrpc/gpyrpc.proto",
        "kclvm/spec/gpyrpc/protorpc_wire.proto",
        "kclvm/spec/modfile/modfile.proto",
    ],
    url="https://kcl-lang.io/",
    packages=setuptools.find_packages(),
    include_package_data=True,
    zip_safe=True,
    install_requires=install_requires,
)



# Copyright 2021 The KCL Authors. All rights reserved.

import sys
from dataclasses import dataclass

import kclvm.compiler.parser as parser
import kclvm.compiler.build.compiler as compiler

from .transformer import transform_ast_to_kclx_ast_json_str

USAGE = """\
usage: kclvm -m kclvm.internal.kclx -f=<file>
usage: kclvm -m kclvm.internal.kclx -h
"""


@dataclass
class CmdFlags:
    help: bool = False
    file: str = ""


def parse_flags(args: list) -> CmdFlags:
    m = CmdFlags()
    for s in args:
        if s == "-h" or s == "-help":
            m.help = True
            continue

        if s.startswith("-f="):
            value = s[len("-f=") :]
            m.file = value
            continue

    return m


def main():
    flags = parse_flags(sys.argv[1:])

    if flags.help:
        print(USAGE)
        sys.exit(0)

    if flags.file:
        ast_prog = parser.LoadProgram(*flags.file.split(","))
        compiler.FixAndResolveProgram(ast_prog)
        print(transform_ast_to_kclx_ast_json_str(ast_prog))


if __name__ == "__main__":
    main()
